import textwrap
from dataclasses import dataclass, field
from typing import List, Optional, Type

import numpy as np

import biofit.config as config
from biofit.integration.biosets import get_feature
from biofit.integration.R import RCaller
from biofit.processing import SelectedColumnTypes, sync_backup_config
from biofit.utils.types import Unset
from biofit.visualization.plotting import BasePlotter, PlotterConfig


def prepare_data_for_hist(x1, x2=None):
    # Calculate row sums and column sums for dataset 1
    data1_sample = x1.sum(axis=1)
    data1_feature = x1.sum(axis=0)

    list_sums = [data1_sample, data1_feature]

    # If dataset 2 is provided, calculate its row and column sums
    if x2 is not None:
        data2_sample = x2.sum(axis=1)
        data2_feature = x2.sum(axis=0)
        list_sums.extend([data2_sample, data2_feature])

    return list_sums


def non_zero_sums(x1, x2=None):
    # Sum all non-zero values for dataset 1
    x1_non_zero_row = (x1 != 0).sum(axis=1)
    x1_non_zero_col = (x1 != 0).sum(axis=0)
    sums = [x1_non_zero_row, x1_non_zero_col]

    # If dataset 2 is provided, calculate its non-zero sums
    if x2 is not None:
        x2_non_zero_row = (x2 != 0).sum(axis=1)
        x2_non_zero_col = (x2 != 0).sum(axis=0)
        sums.extend([x2_non_zero_row, x2_non_zero_col])

    return sums


def prepare_axis_label(label, log_type):
    if "1p" in log_type:
        if "_1p" in log_type:
            label_log = log_type.replace("_1p", "")
            label = f"{label} ({label_log}(x+1))"
        else:
            label = f"{label} (ln(x+1))"
    elif log_type == "log":
        label = f"{label} (ln)"
    else:
        label = f"{label} ({log_type})"
    return label


def log_transformation(x, log_type):
    if "1p" in log_type:
        if "_1p" in log_type:
            label_log = log_type.replace("_1p", "")
            if label_log == "log10":
                return np.log10(1 + x)
            elif label_log == "log2":
                return np.log2(1 + x)
        else:
            return np.log1p(x)
    elif log_type == "log":
        return np.log(x)
    elif log_type == "log2":
        return np.log2(x)
    elif log_type == "log10":
        return np.log10(x)
    return x


@dataclass
class HistogramConfig(PlotterConfig):
    processor_type: str = field(default="scaling", init=False, repr=False)
    _unused_feature_types: List[Type] = field(
        default=get_feature("METADATA_FEATURE_TYPES"),
        init=False,
        repr=False,
    )
    r_source: str = field(
        default=(config.R_SCRIPTS / "plotting_utils.R").as_posix(),
        init=False,
        repr=False,
    )
    main_method: str = field(default="generate_histogram", init=False, repr=False)

    xlab: str = "X"
    ylab: str = "Frequency"
    title: str = "Histogram"
    bins: int = 30
    font_size: int = 8
    col_fill: str = "grey40"
    col_outline: str = "white"
    col_fill = "grey40"
    col_outline = ("black",)
    x1_name: str = "Before"
    x2_name: str = "After"
    xlog: Optional[str] = None
    ylog: Optional[str] = None

    def __post_init__(self):
        if self.xlog not in [
            None,
            "log2",
            "log10",
            "log",
            "log2_1p",
            "log10_1p",
            "log1p",
        ]:
            raise ValueError(
                f"Invalid value for xlog: {self.xlog}. Must be one of: None, 'log2', 'log10', 'log', 'log2_1p', 'log10_1p', 'log1p'"
            )
        if self.ylog not in [
            None,
            "log2",
            "log10",
            "log",
            "log2_1p",
            "log10_1p",
            "log1p",
        ]:
            raise ValueError(
                f"Invalid value for ylog: {self.ylog}. Must be one of: None, 'log2', 'log10', 'log', 'log2_1p', 'log10_1p', 'log1p'"
            )


class HistogramPlotter(BasePlotter):
    _config_class = HistogramConfig
    config: HistogramConfig

    def __init__(
        self,
        xlab: str = Unset('"X"'),
        ylab: str = Unset('"Frequency"'),
        title: str = Unset('"Histogram"'),
        bins: int = Unset("30"),
        font_size: int = Unset("8"),
        col_fill: str = Unset('"grey40"'),
        col_outline: str = Unset('"white"'),
        xlog: Optional[str] = Unset("None"),
        ylog: Optional[str] = Unset("None"),
        install_missing: bool = None,
        config: Optional[HistogramConfig] = None,
    ):
        super().__init__(
            xlab=xlab,
            ylab=ylab,
            title=title,
            bins=bins,
            font_size=font_size,
            col_fill=col_fill,
            col_outline=col_outline,
            xlog=xlog,
            ylog=ylog,
            config=config,
            install_missing=install_missing,
        )
        if self.config.r_source and self.config.main_method:
            self.r_caller = RCaller.from_script(self.config.r_source)
            enter_code = textwrap.dedent(
                """
                suppressPackageStartupMessages(require(ggplot2))
                """
            )
            exit_code = textwrap.dedent(
                """
                ggplot2::ggsave(path, results)
                """
            )
            self.plotter = self.r_caller.get_method(
                self.config.main_method, enter_code, exit_code
            )

    @sync_backup_config
    def set_params(self, **kwargs):
        self.config = self.config.replace_defaults(**kwargs)
        if self.config.r_source and self.config.main_method:
            self.r_caller = RCaller.from_script(self.config.r_source)
            enter_code = textwrap.dedent(
                """
                suppressPackageStartupMessages(require(ggplot2))
                """
            )
            exit_code = textwrap.dedent(
                """
                ggplot2::ggsave(path, results)
                """
            )
            self.plotter = self.r_caller.get_method(
                self.config.main_method, enter_code, exit_code
            )

        return self

    def plot(
        self,
        x,
        input_columns: SelectedColumnTypes = None,
        xlab: str = Unset('"X"'),
        ylab: str = Unset('"Frequency"'),
        title: str = Unset('"Histogram"'),
        bins: int = Unset("30"),
        font_size: int = Unset("8"),
        col_fill: str = Unset('"grey40"'),
        col_outline: str = Unset('"white"'),
        xlog: Optional[str] = Unset("None"),
        ylog: Optional[str] = Unset("None"),
        path: str = None,
        device: str = "pdf",
        fingerprint: str = None,
        unused_columns: SelectedColumnTypes = None,
        raise_if_missing: bool = True,
        show: bool = True,
    ):
        self.config._input_columns = self._set_input_columns_and_arity(input_columns)
        return self._plot(
            x,
            xlab=xlab,
            ylab=ylab,
            title=title,
            bins=bins,
            font_size=font_size,
            col_fill=col_fill,
            col_outline=col_outline,
            xlog=xlog,
            ylog=ylog,
            path=path,
            device=device,
            fingerprint=fingerprint,
            unused_columns=unused_columns,
            raise_if_missing=raise_if_missing,
            show=show,
        )

    def plot_arrow(self, x):
        kwargs = self.config.get_params()
        context_kwargs = {
            "path": kwargs.pop("path", None),
        }
        self.plotter(x, context_kwargs=context_kwargs, **kwargs)


@dataclass
class ComparisonHistogramConfig(PlotterConfig):
    processor_type: str = field(default="scaling", init=False, repr=False)
    _fit_input_feature_types: List[Type] = field(
        default_factory=lambda: [None, None], init=False, repr=False
    )
    _transform_input_feature_types: List[Type] = field(
        default_factory=lambda: [None, None], init=False, repr=False
    )
    _fit_unused_feature_types: List[Type] = field(
        default_factory=lambda: [
            get_feature("METADATA_FEATURE_TYPES"),
            get_feature("METADATA_FEATURE_TYPES"),
        ],
        init=False,
        repr=False,
    )
    _transform_unused_feature_types: List[Type] = field(
        default_factory=lambda: [
            get_feature("METADATA_FEATURE_TYPES"),
            get_feature("METADATA_FEATURE_TYPES"),
        ],
        init=False,
        repr=False,
    )
    r_source: str = field(
        default=(config.R_SCRIPTS / "plotting_utils.R").as_posix(),
        init=False,
        repr=False,
    )
    main_method: str = field(
        default="generate_comparison_histogram", init=False, repr=False
    )

    xlab: Optional[str] = None
    ylab: str = "Count"
    title: str = "Comparison Histogram"
    bins: int = 30
    alpha: float = 0.6
    legend_title: str = "Legend"
    legend_position: str = "top"
    subplot_title1: str = "Before"
    subplot_title2: str = "After"
    col_set: str = "Set1"
    cols: Optional[List[str]] = None
    xlog: Optional[bool] = None
    ylog: Optional[bool] = None


class ComparisonHistogramPlotter(BasePlotter):
    _config_class = ComparisonHistogramConfig
    config: ComparisonHistogramConfig

    def __init__(
        self,
        xlab: Optional[str] = Unset("None"),
        ylab: str = Unset("None"),
        title: str = Unset("None"),
        bins: int = Unset("None"),
        alpha: float = Unset("None"),
        legend_title: str = Unset("None"),
        legend_position: str = Unset("None"),
        subplot_title1: str = Unset("None"),
        subplot_title2: str = Unset("None"),
        col_set: str = Unset("None"),
        cols: Optional[List[str]] = Unset("None"),
        xlog: Optional[bool] = Unset("None"),
        ylog: Optional[bool] = Unset("None"),
        install_missing: bool = None,
        config: Optional[ComparisonHistogramConfig] = None,
    ):
        super().__init__(
            config=config,
            xlab=xlab,
            ylab=ylab,
            title=title,
            bins=bins,
            alpha=alpha,
            legend_title=legend_title,
            legend_position=legend_position,
            subplot_title1=subplot_title1,
            subplot_title2=subplot_title2,
            col_set=col_set,
            cols=cols,
            xlog=xlog,
            ylog=ylog,
            install_missing=install_missing,
        )
        self.plotter = None
        if self.config.r_source and self.config.main_method:
            self.r_caller = RCaller.from_script(self.config.r_source)
            enter_code = textwrap.dedent(
                """
                suppressPackageStartupMessages(require(ggplot2))
                """
            )
            exit_code = textwrap.dedent(
                """
                ggplot2::ggsave(path, results)
                """
            )
            self.plotter = self.r_caller.get_method(
                self.config.main_method, enter_code, exit_code
            )

    @sync_backup_config
    def set_params(self, **kwargs):
        self.config = self.config.replace_defaults(**kwargs)
        if self.config.r_source and self.config.main_method:
            self.r_caller = RCaller.from_script(self.config.r_source)
            enter_code = textwrap.dedent(
                """
                suppressPackageStartupMessages(require(ggplot2))
                """
            )
            exit_code = textwrap.dedent(
                """
                ggplot2::ggsave(path, results)
                """
            )
            self.plotter = self.r_caller.get_method(
                self.config.main_method, enter_code, exit_code
            )

        return self

    def plot(
        self,
        x1,
        x2=None,
        column1: SelectedColumnTypes = None,
        column2: SelectedColumnTypes = None,
        xlab: Optional[str] = Unset("None"),
        ylab: str = Unset("None"),
        title: str = Unset("None"),
        bins: int = Unset("None"),
        alpha: float = Unset("None"),
        legend_title: str = Unset("None"),
        legend_position: str = Unset("None"),
        subplot_title1: str = Unset("None"),
        subplot_title2: str = Unset("None"),
        col_set: str = Unset("None"),
        cols: Optional[List[str]] = Unset("None"),
        xlog: Optional[bool] = Unset("None"),
        ylog: Optional[bool] = Unset("None"),
        path: str = None,
        device: str = "pdf",
        fingerprint: str = None,
        unused_columns: SelectedColumnTypes = None,
        raise_if_missing: bool = True,
        show: bool = True,
    ):
        self.config._input_columns = self._set_input_columns_and_arity(column1, column2)
        return self._plot(
            x1,
            x2,
            xlab=xlab,
            ylab=ylab,
            title=title,
            bins=bins,
            alpha=alpha,
            legend_title=legend_title,
            legend_position=legend_position,
            col_set=col_set,
            cols=cols,
            subplot_title1=subplot_title1,
            subplot_title2=subplot_title2,
            xlog=xlog,
            ylog=ylog,
            path=path,
            device=device,
            fingerprint=fingerprint,
            unused_columns=unused_columns,
            raise_if_missing=raise_if_missing,
            show=show,
        )

    def plot_arrow(self, x1, x2):
        kwargs = self.config.get_params()
        self.plotter(x1, x2, **kwargs)



from dataclasses import dataclass, field
from typing import TYPE_CHECKING, List, Type, Union

import numpy as np
from biocore import DataHandler
from biocore.utils.import_util import is_imblearn_available, requires_backends

from biofit.integration.biosets import get_feature
from biofit.processing import SelectedColumnTypes, sync_backup_config
from biofit.utils import logging

from ..resampling import Resampler, ResamplerConfig

if TYPE_CHECKING:
    from sklearn.base import BaseEstimator

    from biofit.processing import BaseProcessor

logger = logging.get_logger(__name__)

OVER_SAMPLING_METHODS = {}

if is_imblearn_available():
    from imblearn.over_sampling import (
        ADASYN,
        SMOTE,
        SMOTEN,
        SMOTENC,
        SVMSMOTE,
        KMeansSMOTE,
        RandomOverSampler,
    )

    OVER_SAMPLING_METHODS = {
        "random": RandomOverSampler,
        "smote": SMOTE,
        "smoten": SMOTEN,
        "smotenc": SMOTENC,
        "svmsmote": SVMSMOTE,
        "kmeanssmote": KMeansSMOTE,
        "adasyn": ADASYN,
    }


@dataclass
class UpSamplerConfig(ResamplerConfig):
    # process descriptions
    processor_name: str = field(default="upsampling", init=False, repr=False)
    _fit_input_feature_types: List[Type] = field(
        default_factory=lambda: [None, get_feature("TARGET_FEATURE_TYPES")],
        init=False,
        repr=False,
    )
    _fit_unused_feature_types: List[Type] = field(
        default_factory=lambda: [get_feature("METADATA_FEATURE_TYPES"), None],
        init=False,
        repr=False,
    )
    _transform_unused_feature_types: List[Type] = field(
        default_factory=lambda: [get_feature("METADATA_FEATURE_TYPES")],
        init=False,
        repr=False,
    )

    input_columns: List[str] = field(default=None, init=False, repr=False)
    target_column: str = field(default=None, init=False, repr=False)
    method: str = "random"

    sampling_strategy: str = "auto"
    random_state: int = None

    # RandomOverSampler specific attributes
    shrinkage: dict = None

    # BaseSMOTE attributes
    k_neighbors = 5
    n_jobs = None

    # SMOTEN specific attributes
    categorical_encoder: Union["BaseEstimator", "BaseProcessor"] = None

    # SMOTENC specific attributes
    categorical_features: list = None

    # SVMSMOTE specific attributes
    m_neighbors: int = 10
    svm_estimator: Union["BaseEstimator", "BaseProcessor"] = None
    out_step: float = 0.5

    # KMeansSMOTE specific attributes
    kmeans_estimator: Union[int, "BaseEstimator", "BaseProcessor"] = None
    density_exponent: Union[float, str] = "auto"
    cluster_balance_threshold: Union[float, str] = "auto"

    # ADASYN specific attributes
    n_neighbors: int = 5

    def __post_init__(self):
        requires_backends("upsampling", "imblearn")
        if self.random_state is None:
            self.random_state = np.random.randint(0, np.iinfo(np.int32).max)
        if self.method not in OVER_SAMPLING_METHODS:
            raise ValueError(
                f"Invalid method {self.method}. Valid methods are {list(OVER_SAMPLING_METHODS.keys())}"
            )

        self.sampler_kwargs_ = {}
        self.sampler_kwargs_["sampling_strategy"] = self.sampling_strategy
        self.sampler_kwargs_["random_state"] = self.random_state

        if self.method == "random":
            self.sampler_kwargs_["sampling_strategy"] = self.sampling_strategy

        if self.method in [
            "smote",
            "smoten",
            "smotenc",
            "svmsmote",
            "kmeanssmote",
            "adasyn",
        ]:
            self.sampler_kwargs_["n_jobs"] = self.n_jobs

        if self.method in ["smote", "smoten", "smotenc", "svmsmote", "kmeanssmote"]:
            self.sampler_kwargs_["k_neighbors"] = self.k_neighbors

        if self.method in ["smoten", "smotenc"]:
            self.sampler_kwargs_["categorical_encoder"] = self.categorical_encoder

        if self.method == "smotenc":
            self.sampler_kwargs_["categorical_features"] = self.categorical_features

        if self.method == "svmsmote":
            self.sampler_kwargs_["m_neighbors"] = self.m_neighbors
            self.sampler_kwargs_["svm_estimator"] = self.svm_estimator
            self.sampler_kwargs_["out_step"] = self.out_step

        if self.method == "kmeanssmote":
            self.sampler_kwargs_["kmeans_estimator"] = self.kmeans_estimator
            self.sampler_kwargs_["density_exponent"] = self.density_exponent
            self.sampler_kwargs_["cluster_balance_threshold"] = (
                self.cluster_balance_threshold
            )

        if self.method == "adasyn":
            self.sampler_kwargs_["n_neighbors"] = self.n_neighbors


@dataclass
class UpSamplerConfigForMetagenomics(UpSamplerConfig):
    # dataset specific attributes
    _fit_input_feature_types: List[Type] = field(
        default_factory=lambda: [
            (get_feature("Abundance"), get_feature("ReadCount")),
            get_feature("TARGET_FEATURE_TYPES"),
        ],
        init=False,
        repr=False,
    )
    _transform_input_feature_types: List[Type] = field(
        default_factory=lambda: [(get_feature("Abundance"), get_feature("ReadCount"))],
        init=False,
        repr=False,
    )
    dataset_name: str = field(default="metagenomics", init=False, repr=False)


@dataclass
class UpSamplerConfigForOTU(UpSamplerConfig):
    # dataset specific attributes
    _fit_input_feature_types: List[Type] = field(
        default_factory=lambda: [
            get_feature("Abundance"),
            get_feature("TARGET_FEATURE_TYPES"),
        ],
        init=False,
        repr=False,
    )
    _transform_input_feature_types: List[Type] = field(
        default_factory=lambda: [get_feature("Abundance")], init=False, repr=False
    )
    dataset_name: str = field(default="otu", init=False, repr=False)


@dataclass
class UpSamplerConfigForSNP(UpSamplerConfig):
    # dataset specific attributes
    _fit_input_feature_types: List[Type] = field(
        default_factory=lambda: [
            get_feature("GenomicVariant"),
            get_feature("TARGET_FEATURE_TYPES"),
        ],
        init=False,
        repr=False,
    )
    _transform_input_feature_types: List[Type] = field(
        default_factory=lambda: [get_feature("GenomicVariant")], init=False, repr=False
    )
    dataset_name: str = field(default="snp", init=False, repr=False)


class UpSampler(Resampler):
    output_dtype = "float32"

    # config class
    config_class = UpSamplerConfig
    config: UpSamplerConfig

    def __init__(
        self,
        method: str = "random",
        sampling_strategy: str = "auto",
        random_state: int = None,
        shrinkage: dict = None,
        k_neighbors=5,
        n_jobs=None,
        categorical_encoder: Union["BaseEstimator", "BaseProcessor"] = None,
        categorical_features: list = None,
        m_neighbors: int = 10,
        svm_estimator: Union["BaseEstimator", "BaseProcessor"] = None,
        out_step: float = 0.5,
        kmeans_estimator: Union[int, "BaseEstimator", "BaseProcessor"] = None,
        density_exponent: Union[float, str] = "auto",
        cluster_balance_threshold: Union[float, str] = "auto",
        n_neighbors: int = 5,
        config: Union[
            UpSamplerConfig,
            UpSamplerConfigForOTU,
            UpSamplerConfigForSNP,
            UpSamplerConfigForMetagenomics,
        ] = None,
        **kwargs,
    ):
        """
        Upsample the minority class(es) in the dataset.

        Args:
            method (str, 'random'):
                The method to use for oversampling. Possible options are:
                - 'random': Random Over Sampler
                - 'smote': Synthetic Minority Over-sampling Technique
                - 'smoten': SMOTE for numerical features only
                - 'smotenc': SMOTE for numerical and categorical features
                - 'svmsmote': SVM-SMOTE
                - 'kmeanssmote': KMeans-SMOTE
                - 'adasyn': Adaptive Synthetic Sampling Approach for Imbalanced Learning

            sampling_strategy (str, 'auto'):
                The sampling strategy to use for oversampling. Possible options are:
                - 'auto': Automatically resample the minority class(es) to the majority class(es) size.
                - 'all': Resample all classes to the same size.

            random_state (int, *optional*):
                The random state to use for sampling.

            shrinkage (dict, *optional*):
                The shrinkage parameter for RandomOverSampler.

            k_neighbors (int, 5):
                The number of nearest neighbors to use for SMOTE, SMOTEN, SMOTENC, SVMSMOTE, and KMeansSMOTE.

            n_jobs (int, *optional*):
                The number of jobs to use for SMOTE, SMOTEN, SMOTENC, SVMSMOTE, and KMeansSMOTE.

            categorical_encoder (Union[BaseEstimator, BaseProcessor], *optional*):
                The encoder to use for SMOTEN.

            categorical_features (list, *optional*):
                The list of categorical features to use for SMOTENC.

            m_neighbors (int, 10):
                The number of nearest neighbors to use for SVMSMOTE.

            svm_estimator (Union[BaseEstimator, BaseProcessor], *optional*):
                The estimator to use for SVMSMOTE.

            out_step (float, 0.5):
                The outlier step to use for SVMSMOTE.

            kmeans_estimator (Union[int, BaseEstimator, BaseProcessor], *optional*):
                The estimator to use for KMeansSMOTE.

            density_exponent (Union[float, str], 'auto'):
                The exponent to use for KMeansSMOTE.

            cluster_balance_threshold (Union[float, str], 'auto'):
                The balance threshold to use for KMeansSMOTE.

            n_neighbors (int, 5):
                The number of nearest neighbors to use for ADASYN.

            config (Union[UpSamplerConfig, UpSamplerConfigForOTU, UpSamplerConfigForSNP, UpSamplerConfigForMetagenomics], *optional*):
                The configuration to use for the upsampling process. If provided, the other arguments are ignored.
                Use set_params to update the configuration after initialization.
            **kwargs:
                Additional keyword arguments to be passed to ProcessorConfig
        """
        super().__init__(
            config=config,
            method=method,
            sampling_strategy=sampling_strategy,
            random_state=random_state,
            shrinkage=shrinkage,
            k_neighbors=k_neighbors,
            n_jobs=n_jobs,
            categorical_encoder=categorical_encoder,
            categorical_features=categorical_features,
            m_neighbors=m_neighbors,
            svm_estimator=svm_estimator,
            out_step=out_step,
            kmeans_estimator=kmeans_estimator,
            density_exponent=density_exponent,
            cluster_balance_threshold=cluster_balance_threshold,
            n_neighbors=n_neighbors,
            **kwargs,
        )

        self.over_sampler = OVER_SAMPLING_METHODS[self.config.method](
            **self.config.sampler_kwargs_
        )

    @sync_backup_config
    def set_params(self, **kwargs):
        self.config = self.config.replace_defaults(**kwargs)
        self.over_sampler = OVER_SAMPLING_METHODS[self.config.method](
            **self.config.sampler_kwargs_
        )
        return self

    def fit(
        self,
        X,
        y=None,
        input_columns: SelectedColumnTypes = None,
        target_column: SelectedColumnTypes = None,
        raise_if_missing: bool = None,
        cache_output: bool = None,
        load_from_cache_file: bool = None,
        batched: bool = True,
        batch_size: int = 1000,
        batch_format: str = None,
        num_proc: int = None,
        map_kwargs: dict = {"fn_kwargs": {}},
        cache_dir: str = None,
        cache_file_name: str = None,
        fingerprint: str = None,
    ) -> "UpSampler":
        self.config._input_columns = self._set_input_columns_and_arity(
            input_columns, target_column
        )
        return self._process_fit(
            X,
            y,
            raise_if_missing=raise_if_missing,
            cache_output=cache_output,
            load_from_cache_file=load_from_cache_file,
            batched=batched,
            batch_size=batch_size,
            batch_format=batch_format,
            num_proc=num_proc,
            map_kwargs=map_kwargs,
            cache_dir=cache_dir,
            cache_file_name=cache_file_name,
            fingerprint=fingerprint,
        )

    def transform(
        self,
        X,
        input_columns: SelectedColumnTypes = None,
        keep_unused_columns: bool = None,
        raise_if_missing: bool = None,
        cache_output: bool = None,
        cache_dir: str = None,
        cache_file_name: str = None,
        load_from_cache_file: bool = None,
        batched: bool = None,
        batch_size: int = None,
        batch_format: str = None,
        output_format: str = None,
        map_kwargs: dict = None,
        num_proc: int = None,
        fingerprint: str = None,
    ):
        self._input_columns = self._set_input_columns_and_arity(input_columns)
        return self._process_transform(
            X,
            keep_unused_columns=keep_unused_columns,
            raise_if_missing=raise_if_missing,
            cache_output=cache_output,
            cache_dir=cache_dir,
            cache_file_name=cache_file_name,
            load_from_cache_file=load_from_cache_file,
            batched=batched,
            batch_size=batch_size,
            batch_format=batch_format,
            output_format=output_format,
            map_kwargs=map_kwargs,
            num_proc=num_proc,
            fingerprint=fingerprint,
        )

    def fit_transform(
        self,
        X,
        y=None,
        input_columns: SelectedColumnTypes = None,
        target_column: SelectedColumnTypes = None,
        keep_unused_columns: bool = True,
        raise_if_missing: bool = True,
        cache_output: bool = None,
        load_from_cache_file: bool = None,
        batched: bool = True,
        batch_size: int = 1000,
        batch_format: str = None,
        output_format: str = None,
        map_kwargs: dict = {"fn_kwargs": {}},
        num_proc: int = None,
        cache_dir: str = None,
        cache_file_name: str = None,
        fingerprint: str = None,
    ):
        return self.fit(
            X,
            y,
            input_columns=input_columns,
            target_column=target_column,
            raise_if_missing=raise_if_missing,
            cache_output=cache_output,
            load_from_cache_file=load_from_cache_file,
            batched=batched,
            batch_size=batch_size,
            batch_format=batch_format,
            num_proc=num_proc,
            map_kwargs=map_kwargs,
            cache_dir=cache_dir,
            cache_file_name=cache_file_name,
            fingerprint=fingerprint,
        ).transform(
            X,
            input_columns=input_columns,
            keep_unused_columns=keep_unused_columns,
            raise_if_missing=raise_if_missing,
            cache_output=cache_output,
            cache_dir=cache_dir,
            cache_file_name=cache_file_name,
            load_from_cache_file=load_from_cache_file,
            batched=batched,
            batch_size=batch_size,
            batch_format=batch_format,
            output_format=output_format,
            map_kwargs=map_kwargs,
            num_proc=num_proc,
            fingerprint=fingerprint,
        )

    def _fit_pandas(self, X, y):
        self.over_sampler.fit_resample(X, DataHandler.select_column(y, 0))
        self.config.sample_indices = self.over_sampler.sample_indices_
        return self

    def _fit_numpy(self, X, y):
        self.over_sampler.fit_resample(X, DataHandler.select_column(y, 0))
        self.config.sample_indices = self.over_sampler.sample_indices_
        return self

    def _process_transform_input(self, X, **kwargs):
        batch_size = kwargs.get("batch_size", None)
        if batch_size is not None and batch_size < DataHandler.get_shape(X)[0]:
            logger.warning(
                "Upsampling does not support batch processing. Ignoring batched and batch_size parameters."
            )
            kwargs["batch_size"] = None
        self.config.fingerprint = kwargs.get("new_fingerprint", None)
        kwargs["features"] = None
        return X, kwargs

    def _transform_any(self, X):
        return {"indices": self.config.sample_indices.flatten()}



from dataclasses import dataclass, field

from biofit.processing import BaseProcessor, ProcessorConfig


@dataclass
class TransformerConfig(ProcessorConfig):
    """Base class for Transformer processor configuration."""

    processor_type: str = field(default="transformation", init=False, repr=False)


class Transformer(BaseProcessor):
    """Base class for Transformer processors."""

    config_class = TransformerConfig
    config: TransformerConfig



import copy
import os
import sys
from dataclasses import asdict
from typing import TYPE_CHECKING, Callable, Generator, List, Union

import joblib
import numpy as np
import pandas as pd
import pyarrow as pa
import yaml
from biocore.data_handling import DataHandler
from biocore.utils.import_util import (
    is_catboost_available,
    is_lightgbm_available,
    is_plotly_available,
    is_xgboost_available,
)
from biocore.utils.inspect import get_kwargs
from sklearn.ensemble import GradientBoostingClassifier, GradientBoostingRegressor
from sklearn.model_selection import (
    GroupKFold,
    KFold,
    LeaveOneGroupOut,
    StratifiedGroupKFold,
    StratifiedKFold,
    train_test_split,
)
from sklearn.pipeline import Pipeline
from sklearn.svm import SVC, SVR

from biofit.auto.processing_auto import ProcessorPipeline
from biofit.models import (
    LightGBMForClassification,
    LightGBMForRegression,
    RandomForestForClassification,
    RandomForestForRegression,
)
from biofit.models.lasso.lasso import LassoForClassification, LassoForRegression
from biofit.models.models import Model
from biofit.processing import BaseProcessor
from biofit.utils import logging
from biocore.utils.py_util import is_bioset, is_dataset
from biofit.visualization.plotting_utils import plot_feature_importance

if TYPE_CHECKING:
    import optuna

logger = logging.get_logger(__name__)


CLASSIFICATION_TASKS = ["binary_classification", "multiclass_classification"]
REGRESSION_TASKS = ["regression", "multi_regression"]

_MODELS = {
    "binary_classification": {
        "random_forest": RandomForestForClassification,
        "lightgbm": LightGBMForClassification,
        "svm": SVC,
        "gradient_boosting": GradientBoostingClassifier,
        "lasso": LassoForClassification,
    },
    "multiclass_classification": {
        "random_forest": RandomForestForClassification,
        "lightgbm": LightGBMForClassification,
        "svm": SVC,
        "gradient_boosting": GradientBoostingClassifier,
        "lasso": LassoForClassification,
    },
    "regression": {
        "random_forest": RandomForestForRegression,
        "lightgbm": LightGBMForRegression,
        "svm": SVR,
        "gradient_boosting": GradientBoostingRegressor,
        "lasso": LassoForRegression,
    },
    "multi_regression": {
        "random_forest": RandomForestForRegression,
        "lightgbm": LightGBMForRegression,
        "svm": SVR,
        "gradient_boosting": GradientBoostingRegressor,
        "lasso": LassoForRegression,
    },
}

if is_xgboost_available():
    from xgboost import XGBClassifier, XGBRegressor

    _MODELS["binary_classification"]["xgboost"] = XGBClassifier
    _MODELS["multiclass_classification"]["xgboost"] = XGBClassifier
    _MODELS["regression"]["xgboost"] = XGBRegressor
    _MODELS["multi_regression"]["xgboost"] = XGBRegressor

if is_lightgbm_available():
    _MODELS["binary_classification"]["lightgbm"] = LightGBMForClassification
    _MODELS["multiclass_classification"]["lightgbm"] = LightGBMForClassification
    _MODELS["regression"]["lightgbm"] = LightGBMForRegression
    _MODELS["multi_regression"]["lightgbm"] = LightGBMForRegression

if is_catboost_available():
    from catboost import CatBoostClassifier, CatBoostRegressor

    _MODELS["binary_classification"]["catboost"] = CatBoostClassifier
    _MODELS["multiclass_classification"]["catboost"] = CatBoostClassifier
    _MODELS["regression"]["catboost"] = CatBoostRegressor
    _MODELS["multi_regression"]["catboost"] = CatBoostRegressor


_CV = {
    "binary_classification": {
        "stratified_kfold": StratifiedKFold,
        "kfold": KFold,
        "group_kfold": GroupKFold,
        "stratified_group_kfold": StratifiedGroupKFold,
        "logo": LeaveOneGroupOut,
    },
    "multiclass_classification": {
        "stratified_kfold": StratifiedKFold,
        "kfold": KFold,
        "group_kfold": GroupKFold,
        "stratified_group_kfold": StratifiedGroupKFold,
        "logo": LeaveOneGroupOut,
    },
    "regression": {
        "stratified_kfold": StratifiedKFold,
        "kfold": KFold,
    },
    "multi_regression": {
        "stratified_kfold": StratifiedKFold,
        "kfold": KFold,
    },
    "multilabel_classification": {
        "stratified_kfold": StratifiedKFold,
        "kfold": KFold,
    },
}


def _flatten_pipeline(p):
    pipe = []
    if isinstance(p, list):
        for step in p:
            pipe.extend(_flatten_pipeline(step))
    elif isinstance(p, (Pipeline, ProcessorPipeline)):
        for step in p.steps:
            pipe.extend(
                _flatten_pipeline(step[-1] if isinstance(step, tuple) else step)
            )
    else:
        pipe.append(p)
    return pipe


def _get_name_and_params(obj):
    obj_params = None
    obj_name = None
    if obj:
        if isinstance(obj, (ProcessorPipeline, Pipeline, list)):
            if isinstance(obj, (Pipeline, ProcessorPipeline)):
                obj = [step[-1] if isinstance(step, tuple) else step for step in obj]
            obj_params = [
                p.get_params() if isinstance(p, BaseProcessor) else p.__dict__
                for p in obj
            ]
            obj_name = [
                p.config.processor_name
                if isinstance(p, BaseProcessor)
                and getattr(p.config, "processor_name", None)
                else p.__class__.__name__
                for p in obj
            ]
        else:
            obj_params = obj.get_params()
            obj_name = (
                obj.config.processor_name
                if getattr(obj.config, "processor_name", None)
                else obj.__class__.__name__
            )
    return obj_name, obj_params


def _get_processor_info(
    models,
    preprocessors=None,
):
    if isinstance(models, list) and len(models) == 1:
        models = models[0]

    preprocessor_name, preprocessor_params = None, None
    new_preprocessors = preprocessors
    if isinstance(models, Pipeline) and preprocessors is None:
        if len(models.steps) > 1:
            new_preprocessors = _flatten_pipeline(
                [p[-1] if isinstance(p, tuple) else p for p in models.steps[:-1]]
            )
        if isinstance(new_preprocessors, list):
            if len(new_preprocessors) == 1:
                new_preprocessors = new_preprocessors[0]
            else:
                new_preprocessors = ProcessorPipeline(new_preprocessors)

        new_models = (
            models.steps[-1][-1]
            if isinstance(models.steps[-1], tuple)
            else models.steps[-1]
        )
        model_name, model_params = _get_name_and_params(new_models)
        if new_preprocessors is not None:
            preprocessor_name, preprocessor_params = _get_name_and_params(
                new_preprocessors
            )
    elif isinstance(models, list):
        new_models, model_name, model_params = [], [], []
        new_preprocessors, preprocessor_name, preprocessor_params = [], [], []
        for i, m in enumerate(models):
            _preprocessors = None
            if isinstance(preprocessors, list):
                _preprocessors = preprocessors[i]
            if _preprocessors is None:
                if isinstance(m, Pipeline):
                    if len(m.steps) > 1:
                        _preprocessors = _flatten_pipeline(
                            [p[-1] if isinstance(p, tuple) else p for p in m.steps[:-1]]
                        )
                    m = (
                        m.steps[-1][-1]
                        if isinstance(m.steps[-1], tuple)
                        else m.steps[-1]
                    )
            if isinstance(_preprocessors, list):
                if len(_preprocessors) == 1:
                    _preprocessors = _preprocessors[0]
                else:
                    _preprocessors = ProcessorPipeline(_preprocessors)

            _name, _params = _get_name_and_params(m)
            new_models.append(m)
            model_name.append(_name)
            model_params.append(_params)
            if _preprocessors:
                _name, _params = _get_name_and_params(_preprocessors)
                preprocessor_name.append(_name)
                preprocessor_params.append(_params)
                new_preprocessors.append(_preprocessors)
        if len(preprocessor_name) == 0 or len(preprocessor_params) == 0:
            preprocessor_name, preprocessor_params = None, None
    else:
        model_name, model_params = _get_name_and_params(models)
        new_models = models
        if preprocessors is not None:
            new_preprocessors = preprocessors
            preprocessor_name, preprocessor_params = _get_name_and_params(preprocessors)

    return (
        new_models,
        model_name,
        model_params,
        new_preprocessors,
        preprocessor_name,
        preprocessor_params,
    )


def save_model_and_preprocessor(
    output_dir,
    models,
    preprocessors=None,
):
    if models is not None:
        joblib.dump(models, os.path.join(output_dir, "model.joblib"))

    if preprocessors is not None:
        if isinstance(preprocessors, list):
            preprocessors = ProcessorPipeline(preprocessors)
        joblib.dump(preprocessors, os.path.join(output_dir, "preprocessor.joblib"))


def save_feature_importances(
    output_dir,
    feat_importances,
    preprocessor,
    data,
    target,
    target_columns,
    save_plot=True,
    label_names=None,
    **kwargs,
):
    feat_importances = feat_importances.reset_index(names=["features"])
    if save_plot:
        transformed_dataset = (
            preprocessor.fit_transform(data, load_from_cache_file=False)
            if preprocessor
            else data
        )
        if kwargs.get("input_columns") is not None:
            trans_columns = set(
                DataHandler.get_column_names(transformed_dataset, generate_cols=True)
            )
            kwargs["input_columns"] = [
                col for col in kwargs["input_columns"] if col in trans_columns
            ]

        if is_bioset(data) or is_dataset(data):
            y = DataHandler.select_columns(target, columns=target_columns)
        else:
            transformed_dataset = DataHandler.to_pandas(transformed_dataset)
            y = DataHandler.to_pandas(target, columns=target_columns)

        plot_dir = os.path.join(output_dir, "plots")
        params = dict(
            y=y,
            path=plot_dir,
            target_columns=target_columns,
        )
        params.update(kwargs)

        if feat_importances.iloc[:, 1:].sum().sum() == 0:
            logger.warning(
                "All feature importances are zero. Please check the model and data. "
                "Skipping feature importance plot."
            )
        else:
            path = params.pop("path", None)
            params["output_dir"] = path
            plot_feature_importance(
                X=transformed_dataset,
                feature_importances=feat_importances,
                label_names=label_names,
                **params,
            )

    feature_metadata = None
    if "feature_metadata" in kwargs:
        feature_metadata = kwargs["feature_metadata"]
    elif is_bioset(data):
        from biosets import get_feature_metadata

        feature_metadata = get_feature_metadata(data)

    if feature_metadata is not None:
        if isinstance(feature_metadata, dict):
            feature_metadata = pd.DataFrame(
                list(feature_metadata.values()), index=list(feature_metadata.keys())
            )
            feature_metadata = feature_metadata.reset_index(names=["features"])
        if feat_importances.shape[1] > 2:
            feat_importances["median"] = feat_importances.iloc[:, 1:].median(axis=1)
        feat_importances = feature_metadata.merge(
            feat_importances, how="inner", on="features"
        )
    feat_importances.to_csv(
        os.path.join(output_dir, "feature_importances.csv"), index=False
    )


def save_study(
    output_dir,
    study,
):
    from optuna.visualization import (
        plot_param_importances,
        plot_slice,
    )

    joblib.dump(study, os.path.join(output_dir, "study.joblib"))
    trials_df = study.trials_dataframe()
    best_trial_col = [""] * len(trials_df)
    best_trial_col[study.best_trial.number] = "Best Trial"
    trials_df["Best Trial"] = best_trial_col
    trials_df.to_csv(os.path.join(output_dir, "trials.csv"), index=False)
    num_trials = len(study.trials)

    plot_dir = os.path.join(output_dir, "plots")
    param_importances_path = os.path.join(plot_dir, "param_importances_plot.pdf")
    slice_path = os.path.join(plot_dir, "slice_plot.pdf")

    os.makedirs(plot_dir, exist_ok=True)

    if is_plotly_available():
        try:
            if num_trials is not None and num_trials > 1:
                plot_param_importances(study).write_image(
                    param_importances_path, format="pdf", engine="kaleido"
                )
            plot_slice(study).write_image(slice_path, format="pdf", engine="kaleido")
        except (ValueError, RuntimeError) as e:
            logger.warning(f"Failed to save optimization plots: {e}")
    else:
        logger.warning_once(
            "Plotly is not installed. Optimization plots will not be saved."
        )


def save_metrics(
    output_dir,
    metrics,
):
    if isinstance(metrics, list) and len(metrics) == 1:
        metrics = pd.DataFrame({k: [v] for k, v in metrics[0].items()})
    elif isinstance(metrics, dict):
        metrics = pd.DataFrame({k: [v] for k, v in metrics.items()})
    metrics.to_csv(os.path.join(output_dir, "metrics.csv"), index=False)


def save_confusion_matrix(output_dir, confusion_matrices, label_names=None):
    if not isinstance(confusion_matrices, list) or len(confusion_matrices) == 1:
        if isinstance(confusion_matrices, list):
            confusion_matrices = confusion_matrices[0]
        confusion_matrices.to_csv(os.path.join(output_dir, "confusion_matrix.csv"))
    else:
        if label_names is None:
            label_names = [f"label_{i}" for i in range(len(confusion_matrices))]
        for i, cm in enumerate(confusion_matrices):
            cm.to_csv(
                os.path.join(output_dir, f"confusion_matrices_{label_names[i]}.csv")
            )


def save_predictions(
    output_dir,
    preds,
    data,
    target=None,
    label_names=None,
    target_columns=None,
    index=None,
):
    # save as csv
    if index is None:
        index = range(len(data))

    if label_names is None:
        if is_bioset(data) or is_dataset(data):
            if target_columns is not None:
                if isinstance(target_columns, list):
                    target_columns = target_columns[0]
                if target_columns in data._info.features:
                    label_names = data._info.features[target_columns].names
        if label_names is None and (is_bioset(target) or is_dataset(target)):
            target_columns = target_columns or DataHandler.get_column_names(target)[0]
            if target_columns is not None:
                if isinstance(target_columns, list):
                    target_columns = target_columns[0]
                if target_columns in target._info.features:
                    label_names = target._info.features[target_columns].names

    if is_bioset(data) or is_dataset(data):
        from biosets.features import Sample

        names = [
            name
            for name, feat in data._info.features.items()
            if isinstance(feat, Sample)
        ]
        if len(names) == 1:
            names = names[0]
            sample_ids = DataHandler.to_numpy(data, names).flatten()

            if sample_ids is not None:
                if index is not None and isinstance(index, (list, np.ndarray)):
                    sample_ids = np.array(sample_ids)[np.array(index)]
                if isinstance(preds.index[0], int):
                    preds.index = [sample_ids[i] for i in preds.index]
                else:
                    preds.index = sample_ids
                preds.index.name = names

    if label_names is not None:
        preds["predicted"] = DataHandler.argmax(
            preds.iloc[:, : len(label_names)], axis=1
        )
        if target is not None:
            preds["actual"] = DataHandler.to_numpy(target).flatten()[index].astype(int)
            preds["actual"] = preds["actual"].map(
                {i: name for i, name in enumerate(label_names)}
            )
            col_names = label_names + ["predicted", "actual"]

            if is_bioset(target) or is_dataset(target):
                from biosets.features import BinClassLabel

                feat = target._info.features[target_columns]
                if isinstance(feat, BinClassLabel) and (
                    feat.positive_labels is not None or feat.negative_labels is not None
                ):
                    if feat.id in DataHandler.get_column_names(data):
                        preds["actual_original"] = DataHandler.to_pandas(
                            data, feat.id
                        ).values.flatten()[index]
                        col_names.append("actual_original")
        else:
            col_names = label_names + ["predicted"]
        preds["predicted"] = preds["predicted"].map(
            {i: name for i, name in enumerate(label_names)}
        )
        preds.columns = col_names
    else:
        preds["predicted"] = DataHandler.argmax(preds, axis=1)
        if target is not None:
            preds["actual"] = (
                DataHandler.to_pandas(target).iloc[index, 0].values.astype(int)
            )

    if preds.index.name is not None:
        preds.to_csv(os.path.join(output_dir, "predictions.csv"), index=True)
    else:
        preds.to_csv(os.path.join(output_dir, "predictions.csv"), index=False)


def _save_train_results(
    data,
    orig_target,
    target,
    output_dir,
    feature_importances=None,
    confusion_matrices=None,
    study: "optuna.Study" = None,
    num_trials=None,
    time_limit=None,
    models=None,
    metrics=None,
    preds=None,
    input_columns: List[str] = None,
    target_columns: List[str] = None,
    label_names: List[str] = None,
    cv: Union[KFold, StratifiedKFold, GroupKFold, StratifiedGroupKFold] = None,
    group_name: str = None,
    outer_cv: Union[KFold, StratifiedKFold, GroupKFold, StratifiedGroupKFold] = None,
    outer_group_name: str = None,
    task: str = None,
    eval_metric: Union[str, Callable] = None,
    random_state: Union[int, List[int]] = None,
    use_suggested_hyperparameters: bool = True,
    unused_columns: List[str] = None,
    valid_split: str = None,
    index: List[int] = None,
    feature_importance_plot_params=None,
    **kwargs,
):
    orig_data = data
    data, target, _, _, _, _, _, target_columns = _get_data(
        data=data,
        target=target,
        valid_data=None,
        valid_target=None,
        groups=None,
        group_name=None,
        input_columns=input_columns,
        target_columns=target_columns,
        format=None,
        target_required=True,
    )

    if isinstance(preds, list) and len(preds) == 1:
        preds = preds[0]

    os.makedirs(output_dir, exist_ok=True)
    if study:
        save_study(output_dir, study)

    preprocessors = None
    if models is not None:
        (
            models,
            model_name,
            model_params,
            preprocessors,
            preprocessor_name,
            preprocessor_params,
        ) = _get_processor_info(models)
        save_model_and_preprocessor(output_dir, models, preprocessors)

    if metrics is not None:
        save_metrics(output_dir, metrics)

    if preds is not None:
        save_predictions(
            output_dir,
            preds,
            orig_data,
            target if orig_target is None else orig_target,
            label_names,
            target_columns,
            index,
        )

    if confusion_matrices is not None:
        save_confusion_matrix(output_dir, confusion_matrices)

    if feature_importances is not None:
        feature_importance_plot_params = feature_importance_plot_params or {}
        save_feature_importances(
            output_dir,
            feature_importances,
            preprocessors,
            orig_data,
            target,
            target_columns,
            input_columns=input_columns,
            label_names=label_names,
            **feature_importance_plot_params,
        )

    params = {
        "preprocessing": [],
        "optimization": {},
        "training": {},
        "cross_validation": {},
        "model": {},
        "misc": {},
    }

    # General Information
    if task is not None:
        params["training"]["task"] = task

    if label_names is not None:
        if not isinstance(target_columns, list):
            tc = [target_columns]
        else:
            tc = target_columns

        if not isinstance(label_names, (np.ndarray, list)) or not isinstance(
            label_names[0], (np.ndarray, list)
        ):
            label_names = [label_names]

        if "classification" in task:
            target_type = "labels"
        else:
            target_type = "targets"

        params["training"][target_type] = {}
        for i, lab_name in enumerate(tc):
            if lab_name is not None:
                class_feat = None
                if (
                    is_bioset(target) or is_dataset(target)
                ) and lab_name in target._info.features:
                    class_feat = target._info.features[lab_name]
                elif is_bioset(data) or is_dataset(data):
                    class_feat = data._info.features[lab_name]
                if class_feat is not None:
                    if "biosets" in sys.modules and isinstance(
                        class_feat, sys.modules["biosets"].BinClassLabel
                    ):
                        _names = class_feat.names
                        positive_labels = class_feat.positive_labels
                        negative_labels = class_feat.negative_labels
                        if positive_labels is not None or negative_labels is not None:
                            params["training"][target_type][lab_name] = {
                                _names[0]: negative_labels,
                                _names[1]: positive_labels,
                            }
                        else:
                            params["training"][target_type][lab_name] = _names
                    else:
                        lab_name = lab_name or f"{i + 1}"
                        params["training"][target_type][lab_name] = class_feat.names
            else:
                lab_name = lab_name or f"{i + 1}"
                params["training"][target_type][lab_name] = DataHandler.to_list(
                    label_names[i]
                )

        if len(tc) == 1:
            if tc[0] is None or tc[0] not in params["training"][target_type]:
                p = params["training"][target_type]
                if len(p):
                    params["training"][target_type] = params["training"][target_type][
                        next(iter(p))
                    ]
            else:
                params["training"][target_type] = params["training"][target_type][tc[0]]

    if target_columns is not None:
        params["training"]["target_columns"] = target_columns
    if eval_metric is not None:
        params["training"]["eval_metric"] = (
            eval_metric.__name__ if callable(eval_metric) else eval_metric
        )
    if random_state is not None:
        params["training"]["random_state"] = random_state

    if valid_split is not None:
        params["training"]["valid_split"] = valid_split

    # Hyperparameter Tuning Information
    if num_trials is not None:
        params["optimization"]["num_trials"] = num_trials
    if time_limit is not None:
        params["optimization"]["time_limit"] = time_limit
    # list tuned hyperparameters and their ranges
    if study is not None:
        search_space = study.get_trials()[0].distributions
        params["optimization"]["search_space"] = {}
        for param_name, dist in search_space.items():
            dist_pars = dist.__dict__
            dist_pars.pop("step", None)
            params["optimization"]["search_space"][param_name] = {
                "type": (dist.__class__.__name__),
                **dist_pars,
            }

    # Cross-Validation Information
    if cv is not None:
        params["cross_validation"]["cv"] = {
            "type": cv.__class__.__name__,
            "params": cv.__dict__,
        }

    if outer_cv is not None:
        params["cross_validation"]["outer_cv"] = {
            "type": outer_cv.__class__.__name__,
            "params": outer_cv.__dict__,
        }

    # Model Information
    if model_name is not None:
        params["model"]["name"] = model_name
        params["model"]["params"] = model_params if model_params is not None else {}

    # Preprocessor Information
    if preprocessor_name is not None and preprocessor_params is not None:
        if isinstance(preprocessor_name, list):
            params["preprocessing"] = [
                {"name": name, "params": params}
                for name, params in zip(preprocessor_name, preprocessor_params)
            ]
        else:
            params["preprocessing"] = {
                "name": preprocessor_name,
                "params": preprocessor_params,
            }

    if unused_columns is not None:
        params["misc"]["output_dir"] = output_dir

    if unused_columns is not None:
        params["misc"]["unused_columns"] = unused_columns

    if outer_group_name is not None:
        params["misc"]["outer_group_name"] = outer_group_name

    if group_name is not None:
        params["misc"]["group_name"] = group_name

    with open(os.path.join(output_dir, "training_params.yaml"), "w") as f:
        # uncomment for debugging
        # print(yaml.safe_dump(params, sort_keys=False))
        yaml.safe_dump(params, f, sort_keys=False)


def _init_cv(
    cv, groups=None, cv_params=None, sub_task="classification", shuffle=True, seed=42
):
    n_splits = 5
    if isinstance(cv, int):
        n_splits = cv
        if groups is not None:
            cv = LeaveOneGroupOut
        else:
            cv = next(iter(_CV[sub_task].values()))

    if isinstance(cv, str):
        cv = _CV[sub_task][cv]

    if isinstance(cv, type):
        if cv_params is None:
            if cv.__module__.split(".")[0] in ["biofit", "sklearn"]:
                cv_params = {}
                cv_params["shuffle"] = shuffle
                cv_params["random_state"] = seed if shuffle else None
                cv_params["n_splits"] = n_splits
            else:
                raise ValueError(
                    "Please provide cv_params for custom cross validation."
                )
        cv_params = get_kwargs(cv_params, cv)
        cv = cv(**cv_params)
    return cv


def get_task(models):
    if isinstance(models, list) and len(models):
        model = models[0]
    else:
        model = models
    if isinstance(model, (ProcessorPipeline, Pipeline)):
        for m in model.steps:
            val = m
            if isinstance(val, tuple):
                val = val[-1]
            if hasattr(val, "config") and hasattr(val.config, "class_names"):
                return val.config.class_names

    elif hasattr(model, "config") and hasattr(model.config, "class_names"):
        return model.config.class_names
    return None


def set_class_names(models, class_names):
    if class_names is None:
        return models
    if isinstance(models, list) and len(models):
        return [set_class_names(m, class_names) for m in models]
    if isinstance(models, (ProcessorPipeline, Pipeline)):
        for val in models.steps:
            if isinstance(val, tuple) and hasattr(val[-1], "config"):
                val[-1].config.class_names = class_names
            if hasattr(val, "config"):
                val.config.class_names = class_names
    elif hasattr(models, "config"):
        models.config.class_names = class_names
    return models


def infer_task(data=None, target=None, target_columns=None, task=None):
    sub_task = None
    if target is None and data is not None:
        _, target, _, _, _, _, _, target_columns = _get_data(
            data=data,
            target=target,
            valid_data=None,
            valid_target=None,
            groups=None,
            group_name=None,
            input_columns=None,
            target_columns=target_columns,
            format=None,
            target_required=False,
        )
    if target is None and task is None:
        raise ValueError("Target is required to infer task.")

    target_dims = DataHandler.get_shape(target) if target is not None else []
    class_names = None
    if task is None and (is_bioset(target) or is_dataset(target)):
        from biosets.features import BinClassLabel, ClassLabel, RegressionTarget

        if isinstance(target_columns, list):
            target_columns = target_columns[0]
        if target_columns and target_columns in target._info.features:
            feat = target._info.features[target_columns]
        else:
            feat = [
                feat
                for feat in target._info.features.values()
                if isinstance(feat, (BinClassLabel, ClassLabel, RegressionTarget))
            ][0]
        if isinstance(feat, (BinClassLabel, ClassLabel)):
            class_names = target._info.features[target_columns].names
            task = "classification"
        elif isinstance(feat, RegressionTarget):
            task = "regression"
    if task == "classification":
        if len(target_dims) == 1 or target_dims[1] == 1:
            n_classes = None
            if target_columns and (is_bioset(target) or is_dataset(target)):
                if isinstance(target_columns, list):
                    target_columns = target_columns[0]
                class_names = target._info.features[target_columns].names
                n_classes = len(class_names)
            else:
                n_classes = DataHandler.nunique(target)

            if n_classes > 2:
                sub_task = "multiclass_classification"
            else:
                sub_task = "binary_classification"
        else:
            sub_task = "multilabel_classification"
    elif task == "regression":
        if len(target_dims) > 1 and target_dims[1] > 1:
            sub_task = "multiregression"
        else:
            sub_task = "regression"
    elif task in CLASSIFICATION_TASKS or task in REGRESSION_TASKS:
        sub_task = task
    else:
        raise ValueError(
            "Invalid task. Please specify either a task, such as 'classification' or "
            f"'regression', or a sub task, such as {CLASSIFICATION_TASKS + REGRESSION_TASKS}."
        )
    return sub_task, class_names


def _iter_preprocessor(preprocessor, condition=None):
    if isinstance(preprocessor, (Pipeline, ProcessorPipeline)):
        start = False
        for n, p in preprocessor.steps:
            if isinstance(p, BaseProcessor):
                if condition == "sample independent":
                    if p.has_fit:
                        yield p.config.processor_name, p
                    else:
                        break
                elif condition == "sample dependent":
                    if not p.has_fit:
                        start = True
                        yield p
                    elif start:
                        yield p
                else:
                    yield p
            else:
                yield p
    elif isinstance(preprocessor, list):
        for p in preprocessor:
            yield p
    else:
        yield preprocessor


def preprocess(
    preprocessor,
    x_train,
    y_train=None,
    x_valid=None,
    y_valid=None,
    cache_dir=None,
    transform_only=False,
    raise_error=False,
):
    try:
        if isinstance(preprocessor, (Pipeline, ProcessorPipeline)):
            for proc in preprocessor.steps:
                p = proc[-1] if isinstance(proc, tuple) else proc
                if isinstance(p, BaseProcessor):
                    extra_kwargs = {
                        "cache_dir": cache_dir,
                        "load_from_cache_file": False,
                    }
                else:
                    extra_kwargs = {}
                if transform_only:
                    x_train = p.transform(x_train, **extra_kwargs)
                else:
                    x_train = p.fit_transform(x_train, **extra_kwargs)
                if x_valid is not None:
                    x_valid = p.transform(x_valid, **extra_kwargs)
        else:
            if isinstance(preprocessor, BaseProcessor):
                extra_kwargs = {"cache_dir": cache_dir, "load_from_cache_file": False}
            else:
                extra_kwargs = {}
            if transform_only:
                x_train = preprocessor.transform(x_train, **extra_kwargs)
            else:
                x_train = preprocessor.fit_transform(x_train, **extra_kwargs)
            if x_valid is not None:
                x_valid = preprocessor.transform(x_valid, **extra_kwargs)
    except ValueError:
        if raise_error:
            raise
        else:
            logger.info("Preprocessing failed, using nan_to_num")
            _train_cols = x_train.columns.tolist()
            x_train = np.nan_to_num(x_train)
            # convert back to dataframe
            x_train = pd.DataFrame(x_train, columns=_train_cols)
            if x_valid is not None:
                _valid_cols = x_valid.columns.tolist()
                x_valid = np.nan_to_num(x_valid)
                x_valid = pd.DataFrame(x_valid, columns=_valid_cols)
            return preprocess(
                preprocessor,
                x_train,
                y_train,
                x_valid,
                y_valid,
                cache_dir,
                transform_only,
                True,
            )

    return x_train, y_train, x_valid, y_valid


def split(cv, X, y=None, groups=None, indices=None):
    if cv is None:
        return [(None, None)]
    # check if cv is a generator
    if isinstance(cv, Generator):
        return cv
    if indices is not None:
        return cv.split(
            DataHandler.select_rows(X, indices),
            y=DataHandler.select_column(DataHandler.select_rows(y, indices), 0)
            if y is not None
            else None,
            groups=DataHandler.select_column(
                DataHandler.select_rows(groups, indices), 0
            )
            if groups is not None
            else None,
        )
    return cv.split(
        X,
        y=DataHandler.select_column(y, 0) if y is not None else None,
        groups=DataHandler.select_column(groups, 0) if groups is not None else None,
    )


def get_model(models_or_name, task=None):
    if isinstance(models_or_name, list) and len(models_or_name):
        mon = models_or_name[0]
    else:
        mon = models_or_name

    if mon is None:
        return None

    model = None
    model_cls = None
    if isinstance(mon, str):
        model_name = mon
        model_cls = _MODELS[task][model_name]
    elif isinstance(mon, type):
        model_cls = mon
        model_name = mon.__module__.split(".")[-1]
    elif isinstance(mon, Model):
        model_name = mon.config.processor_name
        model_cls = model.__class__
        model = mon
    elif isinstance(mon, (ProcessorPipeline, Pipeline)):
        for m in _flatten_pipeline(mon):
            out = get_model(m, task)
            if out is not None:
                return out
        return None
    elif hasattr(mon, "predict"):
        model_cls = mon.__class__
        model_name = mon.__class__.__module__.split(".")[-1]
        model = mon
    else:
        return None
    return model, model_cls, model_name


def get_model_info(model, task=None):
    out = get_model(model, task=None)
    if out is not None:
        model, _, _ = out
        if hasattr(model, "config"):
            return asdict(model.config)
        elif hasattr(model, "get_params"):
            return model.get_params()
        else:
            return model.__dict__

    return {}


def _split_data(
    data, target, valid_split, seed, shuffle, stratify_by_column, keep_in_memory=True
):
    def _convert_to_in_memory(data, indices=None):
        from datasets import InMemoryTable, MemoryMappedTable

        try:
            if data._indices or indices is not None:
                if indices is None:
                    indices = data._indices.column("indices")

                if isinstance(data._data, MemoryMappedTable):
                    table = MemoryMappedTable._apply_replays(
                        data._data.table, data._data.replays
                    )
                else:
                    table = data._data.table
                table = table.take(indices)
                data._indices = None
                data._data = InMemoryTable(table)

                return data
        except pa.ArrowInvalid:
            logger.error(
                "Table is too large for row selection. Please set keep_in_memory=False or "
                "shuffle=False until https://github.com/apache/arrow/issues/25822 has been resolved"
            )
            raise

    if is_bioset(data) or is_dataset(data):
        if keep_in_memory:
            _convert_to_in_memory(data)

        train_data, valid_data = data.train_test_split(
            test_size=valid_split,
            seed=seed if shuffle else None,
            shuffle=shuffle,
            stratify_by_column=stratify_by_column if shuffle else None,
        ).values()
        if not shuffle and not stratify_by_column:
            target, valid_target = target.train_test_split(
                test_size=valid_split,
                seed=seed,
                shuffle=shuffle,
                stratify_by_column=stratify_by_column,
            )
        else:
            train_indices = train_data._indices.column("indices")
            if valid_data._indices:
                valid_indices = valid_data._indices.column("indices")
            else:
                valid_indices = pa.array(
                    list(
                        sorted(
                            set(range(train_data.num_rows))
                            - set(DataHandler.to_numpy(train_indices).tolist())
                        )
                    )
                )

            if keep_in_memory:
                train_data = _convert_to_in_memory(train_data, train_indices)
                valid_data = _convert_to_in_memory(valid_data, valid_indices)
                target = _convert_to_in_memory(target, train_indices)
                valid_target = _convert_to_in_memory(valid_target, valid_indices)
            else:
                valid_target = copy.deepcopy(target)
                valid_target = valid_target.select(valid_indices)
                target = target.select(train_indices)
    else:
        if stratify_by_column is not None:
            stratify = DataHandler.to_numpy(data, stratify_by_column)
        else:
            stratify = target
        train_data, valid_data, target, valid_target = train_test_split(
            data,
            target,
            test_size=valid_split,
            random_state=seed,
            shuffle=shuffle,
            stratify=stratify,
        )
    return train_data, valid_data, target, valid_target


def _get_label_names(data, target, target_columns, u_labels=None):
    if isinstance(target_columns, list):
        tc = target_columns[0]
    else:
        tc = target_columns
    if (is_bioset(data) or is_dataset(data)) and tc in data._info.features:
        labels = data._info.features[tc].names
    elif (is_bioset(target) or is_dataset(target)) and tc in target._info.features:
        labels = target._info.features[tc].names
    else:
        labels = DataHandler.to_list(u_labels)
    return labels


def _get_data(
    data,
    target=None,
    valid_data=None,
    valid_target=None,
    valid_split=None,
    groups=None,
    outer_groups=None,
    group_name=None,
    outer_group_name=None,
    input_columns="auto",
    target_columns="auto",
    valid_input_columns="auto",
    valid_target_columns="auto",
    target_required=True,
    shuffle=True,
    seed=42,
    keep_in_memory=True,
    format="pandas",
):
    if DataHandler.supports_named_columns(data):
        use_auto_target_columns = target_columns == "auto"
    else:
        input_columns = None if input_columns == "auto" else input_columns
        target_columns = None if target_columns == "auto" else target_columns

    if valid_data is not None and not DataHandler.supports_named_columns(valid_data):
        valid_input_columns = (
            None if valid_input_columns == "auto" else valid_input_columns
        )
        valid_target_columns = (
            None if valid_target_columns == "auto" else valid_target_columns
        )

    def get_target_if_none(data, target_columns):
        target = None
        if target_columns is not None:
            sel_kwargs = {}
            if is_bioset(data) or is_dataset(data):
                sel_kwargs = {"keep_old_fingerprint": False}
            if isinstance(target_columns, str):
                target = DataHandler.select_columns(
                    data, [target_columns], **sel_kwargs
                )
            else:
                target = DataHandler.select_columns(data, target_columns, **sel_kwargs)
        return target

    def handle_target_columns(data, target_columns, required=True):
        if target_columns == "auto":
            if is_bioset(data):
                from biosets import get_target_col_names

                target_columns = get_target_col_names(data)
            elif target_required:
                ValueError(
                    "`data` must be a `Dataset` to automatically infer target columns. "
                    "Please provide `target_columns` or `target`."
                )
        if isinstance(target_columns, (type, tuple)):
            if is_bioset(data) or is_dataset(data):
                target_columns = [
                    k
                    for k, v in data._info.features.items()
                    if isinstance(v, target_columns)
                ]
        if target_columns == "auto":
            target_columns = None
        if target_columns is None and target_required:
            raise ValueError("Target columns must be provided if target is None.")

        if isinstance(target_columns, list) and len(target_columns) == 1:
            target_columns = target_columns[0]
        return target_columns

    def handle_input_columns(data, input_columns):
        if is_bioset(data) or is_dataset(data):
            if input_columns == "auto" and is_bioset(data):
                from biosets import get_data_col_names

                input_columns = get_data_col_names(data)
            elif isinstance(input_columns, (type, tuple)):
                input_columns = [
                    k
                    for k, v in data._info.features.items()
                    if isinstance(v, input_columns)
                ]
        if input_columns == "auto":
            input_columns = None

        return input_columns

    x_train = None
    y_train = None
    y_valid = None
    x_valid = None

    if isinstance(data, dict):
        if isinstance(valid_split, str):
            valid_data = data.get(valid_split)
        else:
            valid_data = data.get(
                "valid", data.get("test"), data.get("validation"), None
            )
        if len(data) < 3:
            train_split = [
                k
                for k in data.keys()
                if k not in ["valid", "test", "validation", valid_split]
            ]
            if not train_split:
                raise ValueError("Train split not found in dataset.")
            data = data.get(train_split)
        else:
            data = data.get("train", data.get("training"))

    if target is None:
        target_columns = handle_target_columns(
            data, target_columns, required=target_required
        )
        y_train = get_target_if_none(data, target_columns)

    else:
        y_train = target
        if DataHandler.supports_named_columns(target):
            target_columns = DataHandler.get_column_names(target)[0]

    if valid_target is None and valid_data is not None:
        if valid_target_columns is None or valid_target_columns == "auto":
            valid_target_columns = handle_target_columns(
                valid_data,
                "auto" if use_auto_target_columns else target_columns,
                required=False,
            )
        if valid_target_columns is not None:
            y_valid = get_target_if_none(valid_data, valid_target_columns)
        else:
            y_valid = get_target_if_none(valid_data, target_columns)
        if valid_target_columns is not None and target_columns is not None:
            if valid_target_columns != target_columns:
                valid_target = DataHandler.set_column_names(
                    valid_target, target_columns
                )
        elif valid_target_columns is not None:
            target_columns = valid_target_columns

    elif valid_target is not None:
        y_valid = valid_target

    if group_name is not None and groups is None:
        groups = DataHandler.select_column(data, group_name)
    if outer_group_name is not None and outer_groups is None:
        outer_groups = DataHandler.select_column(data, outer_group_name)
    if (
        group_name is None
        and groups is not None
        and DataHandler.supports_named_columns(groups)
    ):
        group_name = DataHandler.get_column_names(groups)[0]
    if (
        outer_group_name is None
        and outer_groups is not None
        and DataHandler.supports_named_columns(outer_groups)
    ):
        outer_group_name = DataHandler.get_column_names(outer_groups)[0]

    if valid_data is None and valid_split is not None:
        data, valid_data, y_train, y_valid = _split_data(
            data,
            y_train,
            valid_split,
            seed=seed,
            shuffle=shuffle,
            stratify_by_column=group_name,
            keep_in_memory=keep_in_memory,
        )

    input_columns = handle_input_columns(data, input_columns)
    if valid_data is not None:
        valid_input_columns = handle_input_columns(valid_data, valid_input_columns)
    if input_columns is not None:
        sel_kwargs = {}
        if is_bioset(data) or is_dataset(data):
            sel_kwargs = {"keep_old_fingerprint": False}
        x_train = DataHandler.select_columns(data, input_columns, **sel_kwargs)
        if valid_data is not None:
            sel_kwargs = {}
            if is_bioset(valid_data) or is_dataset(valid_data):
                sel_kwargs = {"keep_old_fingerprint": False}
            valid_input_columns = valid_input_columns or input_columns
            x_valid = DataHandler.select_columns(
                valid_data, valid_input_columns, **sel_kwargs
            )
    elif (
        target is not None
        and DataHandler.supports_named_columns(data)
        and DataHandler.supports_named_columns(target)
    ):
        _input_columns = DataHandler.get_column_names(data)
        _target_columns = DataHandler.get_column_names(target)
        if group_name is not None:
            _target_columns.append(group_name)
        if outer_group_name is not None:
            _target_columns.append(outer_group_name)
        _intersecting_cols = set(_input_columns) & set(_target_columns)
        if _intersecting_cols:
            x_train = DataHandler.drop_columns(data, list(_intersecting_cols))
            if (
                valid_data is not None
                and DataHandler.supports_named_columns(valid_data)
                and DataHandler.supports_named_columns(x_train)
            ):
                sel_kwargs = {}
                if is_bioset(valid_data) or is_dataset(valid_data):
                    sel_kwargs = {"keep_old_fingerprint": False}
                x_valid = DataHandler.select_columns(
                    valid_data, DataHandler.get_column_names(x_train), **sel_kwargs
                )
        else:
            x_train = data
            if valid_data is not None:
                x_valid = valid_data
        input_columns = DataHandler.get_column_names(x_train)
    else:
        x_train = data
        if valid_data is not None:
            x_valid = valid_data
        input_columns = DataHandler.get_column_names(x_train)

    if format is not None:
        x_train = DataHandler.to_format(x_train, format)
        if y_train is not None:
            y_train = DataHandler.to_format(y_train, format)
        if groups is not None:
            groups = DataHandler.to_format(groups, format)
        if outer_groups is not None:
            outer_groups = DataHandler.to_format(outer_groups, format)
        if x_valid is not None:
            x_valid = DataHandler.to_format(x_valid, format)
        if y_valid is not None:
            y_valid = DataHandler.to_format(y_valid, format)

    return (
        x_train,
        y_train,
        x_valid,
        y_valid,
        groups,
        outer_groups,
        input_columns,
        target_columns,
    )



import os
import sys
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, List, Optional, Type

import numpy as np
import pandas as pd

from biofit.integration.biosets import get_feature
from biofit.processing import SelectedColumnTypes
from biofit.utils import logging

from ..stat import Stat, StatConfig

if TYPE_CHECKING:
    import polars as pl

logger = logging.get_logger(__name__)


def is_multiprocess_mode():
    return os.getppid() > 1


def _filter_samples_pandas(X: pd.DataFrame, depth: Optional[float] = None):
    """
    SampleFilter samples in a pandas DataFrame based on their presence in the dataset.

    Args:
        X (pd.DataFrame): The input DataFrame containing the samples.
        min_sample_presence (float, optional): The minimum required presence of a sample in the dataset. Defaults to 0.5.
        depth (float, optional): The depth threshold for numeric columns. If specified, counts values less than depth. Defaults to None.

    Returns:
        List[int]: A list of row indices that pass the filtering condition.
    """
    total_missing = X.isnull().sum(axis=1)
    if depth is not None:
        numeric_rows = X.select_dtypes(include=np.number).index
        total_missing[numeric_rows] += (X.loc[numeric_rows] <= depth).sum(axis=1)
    return total_missing.to_frame()


def _filter_samples_numpy(X: np.ndarray, depth: Optional[float] = None):
    total_missing = np.sum(np.isnan(X) | (X is None), axis=1)
    if depth is not None:
        total_missing += np.sum(X <= depth, axis=1)
    return total_missing[:, np.newaxis]


def _filter_samples_polars(X: "pl.DataFrame", depth: Optional[float] = None):
    if "polars" not in sys.modules:
        import polars
    else:
        polars = sys.modules["polars"]
    total_missing = X.with_columns(
        polars.col("*").is_null() | polars.col("*").is_nan()
    ).sum_horizontal()
    if depth is not None:
        total_missing += X.with_columns(polars.col("*") <= depth).sum_horizontal()
    return total_missing.to_frame()


@dataclass
class RowMissingnessStatConfig(StatConfig):
    # process description
    _transform_process_desc: str = field(
        default="Calculating the number of missing values for each sample",
        init=False,
        repr=False,
    )
    _fit_unused_feature_types: List[Type] = field(
        default_factory=lambda: [get_feature("METADATA_FEATURE_TYPES")],
        init=False,
        repr=False,
    )
    _transform_unused_feature_types: List[Type] = field(
        default_factory=lambda: [get_feature("METADATA_FEATURE_TYPES")],
        init=False,
        repr=False,
    )
    processor_name: str = field(default="row_missingness", init=False, repr=False)
    _n_features_out: int = field(default=1, init=False, repr=False)

    # config attributes
    depth: Optional[float] = None


@dataclass
class RowMissingnessStatConfigForOTU(RowMissingnessStatConfig):
    """Computes the sum of each row in the OTUAbundance feature.
    This class is the same as TotalOTUAbundanceStat.
    It is provided for autostat
    """

    # dataset specific attributes
    _fit_input_feature_types: List[Type] = field(
        default_factory=lambda: [get_feature("Abundance")], init=False, repr=False
    )
    _transform_input_feature_types: List[Type] = field(
        default_factory=lambda: [get_feature("Abundance")], init=False, repr=False
    )
    transform_process_name: str = field(
        default="Calculating sample richness", init=False, repr=False
    )
    dataset_name: str = field(default="otu", init=False, repr=False)

    # config attributes
    depth = 100


@dataclass
class RowMissingnessStatConfigForSNP(RowMissingnessStatConfig):
    """Computes the sum of each row in the OTUAbundance feature.
    This class is the same as TotalOTUAbundanceStat.
    It is provided for autostat
    """

    # dataset specific attributes
    _fit_input_feature_types: List[Type] = field(
        default_factory=lambda: [get_feature("GenomicVariant")], init=False, repr=False
    )
    _transform_input_feature_types: List[Type] = field(
        default_factory=lambda: [get_feature("GenomicVariant")], init=False, repr=False
    )
    transform_process_name: str = field(
        default="Calculating sample richness", init=False, repr=False
    )
    dataset_name: str = field(default="snp", init=False, repr=False)

    # config attributes
    depth = 100


@dataclass
class RowMissingnessStatConfigForMetagenomics(RowMissingnessStatConfig):
    """Computes the sum of each row in the OTUAbundance feature.
    This class is the same as TotalOTUAbundanceStat.
    It is provided for autostat
    """

    # dataset specific attributes
    _fit_input_feature_types: List[Type] = field(
        default_factory=lambda: [(get_feature("Abundance"), get_feature("ReadCount"))],
        init=False,
        repr=False,
    )
    _transform_input_feature_types: List[Type] = field(
        default_factory=lambda: [(get_feature("Abundance"), get_feature("ReadCount"))],
        init=False,
        repr=False,
    )
    transform_process_name: str = field(
        default="Calculating sample richness", init=False, repr=False
    )

    # config attributes
    depth: int = 100


class RowMissingnessStat(Stat):
    # feature attributes
    one_to_one_features = False
    n_features_out = 1

    # config attributes
    _config_class = RowMissingnessStatConfig
    config: RowMissingnessStatConfig

    def fit(
        self,
        X,
        input_columns: SelectedColumnTypes = None,
        raise_if_missing: bool = None,
        cache_output: bool = None,
        load_from_cache_file: bool = None,
        batched: bool = True,
        batch_size: int = 1000,
        batch_format: str = None,
        num_proc: int = None,
        map_kwargs: dict = {"fn_kwargs": {}},
        cache_dir: str = None,
        cache_file_name: str = None,
        fingerprint: str = None,
    ) -> "RowMissingnessStat":
        self.config._input_columns = self._set_input_columns_and_arity(input_columns)
        return self._process_fit(
            X,
            raise_if_missing=raise_if_missing,
            cache_output=cache_output,
            load_from_cache_file=load_from_cache_file,
            batched=batched,
            batch_size=batch_size,
            batch_format=batch_format,
            num_proc=num_proc,
            map_kwargs=map_kwargs,
            cache_dir=cache_dir,
            cache_file_name=cache_file_name,
            fingerprint=fingerprint,
        )

    def transform(
        self,
        X,
        input_columns: SelectedColumnTypes = None,
        keep_unused_columns: bool = None,
        raise_if_missing: bool = None,
        cache_output: bool = None,
        cache_dir: str = None,
        cache_file_name: str = None,
        load_from_cache_file: bool = None,
        batched: bool = None,
        batch_size: int = None,
        batch_format: str = None,
        output_format: str = None,
        map_kwargs: dict = None,
        num_proc: int = None,
        fingerprint: str = None,
    ):
        self._input_columns = self._set_input_columns_and_arity(input_columns)
        return self._process_transform(
            X,
            keep_unused_columns=keep_unused_columns,
            raise_if_missing=raise_if_missing,
            cache_output=cache_output,
            cache_dir=cache_dir,
            cache_file_name=cache_file_name,
            load_from_cache_file=load_from_cache_file,
            batched=batched,
            batch_size=batch_size,
            batch_format=batch_format,
            output_format=output_format,
            map_kwargs=map_kwargs,
            num_proc=num_proc,
            fingerprint=fingerprint,
        )

    def fit_transform(
        self,
        X,
        *args,
        input_columns: SelectedColumnTypes = None,
        keep_unused_columns: bool = True,
        raise_if_missing: bool = False,
        cache_output: bool = None,
        load_from_cache_file: bool = None,
        batched: bool = True,
        batch_size: int = 1000,
        batch_format: str = None,
        output_format: str = None,
        map_kwargs: dict = {"fn_kwargs": {}},
        num_proc: int = None,
        cache_dir: str = None,
        cache_file_name: str = None,
        fingerprint: str = None,
    ):
        return self.fit(
            X,
            input_columns=input_columns,
            raise_if_missing=raise_if_missing,
            cache_output=cache_output,
            load_from_cache_file=load_from_cache_file,
            batched=batched,
            batch_size=batch_size,
            batch_format=batch_format,
            num_proc=num_proc,
            map_kwargs=map_kwargs,
            cache_dir=cache_dir,
            cache_file_name=cache_file_name,
            fingerprint=fingerprint,
        ).transform(
            X,
            input_columns=input_columns,
            keep_unused_columns=keep_unused_columns,
            raise_if_missing=raise_if_missing,
            cache_output=cache_output,
            cache_dir=cache_dir,
            cache_file_name=cache_file_name,
            load_from_cache_file=load_from_cache_file,
            batched=batched,
            batch_size=batch_size,
            batch_format=batch_format,
            output_format=output_format,
            map_kwargs=map_kwargs,
            num_proc=num_proc,
            fingerprint=fingerprint,
        )

    def _transform_numpy(self, X: np.ndarray) -> np.ndarray:
        return _filter_samples_numpy(X, self.config.depth)

    def _transform_pandas(self, X: pd.DataFrame) -> pd.DataFrame:
        return _filter_samples_pandas(X, self.config.depth)

    def _transform_polars(self, X: "pl.DataFrame") -> "pl.DataFrame":
        return _filter_samples_polars(X, self.config.depth)



# ruff: noqa
from .distance import (
    DistanceStat,
    DistanceStatConfig,
    DistanceStatConfigForASV,
    DistanceStatConfigForOTU,
    DistanceStatConfigForMetagenomics,
    DistanceStatConfigForReadCount,
)



from dataclasses import dataclass, field
from typing import TYPE_CHECKING, List, Type, Union

import numpy as np
import pandas as pd

from biofit.integration.biosets import get_feature
from biofit.processing import SelectedColumnTypes
from biofit.utils import logging

from ..filtering import SampleFilter, SampleFilterConfig

if TYPE_CHECKING:
    import polars as pl

logger = logging.get_logger(__name__)


@dataclass
class AbundanceSampleFilterConfig(SampleFilterConfig):
    lower_threshold: Union[int, str] = "auto"
    upper_threshold: Union[int, str] = None
    processor_name: str = field(default="row_abundance", init=False, repr=False)
    _fit_unused_feature_types: List[Type] = field(
        default_factory=lambda: [get_feature("METADATA_FEATURE_TYPES")],
        init=False,
        repr=False,
    )
    _transform_unused_feature_types: List[Type] = field(
        default_factory=lambda: [get_feature("METADATA_FEATURE_TYPES")],
        init=False,
        repr=False,
    )


@dataclass
class AbundanceSampleFilterConfigForOTU(AbundanceSampleFilterConfig):
    _fit_input_feature_types: List[Type] = field(
        default_factory=lambda: [get_feature("Abundance")], init=False, repr=False
    )
    _transform_input_feature_types: List[Type] = field(
        default_factory=lambda: [get_feature("Abundance")], init=False, repr=False
    )
    dataset_name: str = field(default="otu", init=False, repr=False)
    lower_threshold: Union[int, str] = "auto"
    upper_threshold: Union[int, str] = "auto"

    def __post_init__(self):
        self._transform_process_desc = f"SampleFiltering out samples with less than {self.lower_threshold} total otu abundance"


class AbundanceSampleFilter(SampleFilter):
    config_class = AbundanceSampleFilterConfig
    config: AbundanceSampleFilterConfig

    IQRs = None

    def fit(
        self,
        X,
        input_columns: SelectedColumnTypes = None,
        raise_if_missing: bool = None,
        cache_output: bool = None,
        load_from_cache_file: bool = None,
        batched: bool = True,
        batch_size: int = 1000,
        batch_format: str = None,
        num_proc: int = None,
        map_kwargs: dict = {"fn_kwargs": {}},
        cache_dir: str = None,
        cache_file_name: str = None,
        fingerprint: str = None,
    ) -> "AbundanceSampleFilter":
        self.config._input_columns = self._set_input_columns_and_arity(input_columns)
        return self._process_fit(
            X,
            raise_if_missing=raise_if_missing,
            cache_output=cache_output,
            load_from_cache_file=load_from_cache_file,
            batched=batched,
            batch_size=batch_size,
            batch_format=batch_format,
            num_proc=num_proc,
            map_kwargs=map_kwargs,
            cache_dir=cache_dir,
            cache_file_name=cache_file_name,
            fingerprint=fingerprint,
        )

    def transform(
        self,
        X,
        input_columns: SelectedColumnTypes = None,
        keep_unused_columns: bool = None,
        raise_if_missing: bool = None,
        cache_output: bool = None,
        cache_dir: str = None,
        cache_file_name: str = None,
        load_from_cache_file: bool = None,
        batched: bool = None,
        batch_size: int = None,
        batch_format: str = None,
        output_format: str = None,
        map_kwargs: dict = None,
        num_proc: int = None,
        fingerprint: str = None,
    ):
        self._input_columns = self._set_input_columns_and_arity(input_columns)
        return self._process_transform(
            X,
            keep_unused_columns=keep_unused_columns,
            raise_if_missing=raise_if_missing,
            cache_output=cache_output,
            cache_dir=cache_dir,
            cache_file_name=cache_file_name,
            load_from_cache_file=load_from_cache_file,
            batched=batched,
            batch_size=batch_size,
            batch_format=batch_format,
            output_format=output_format,
            map_kwargs=map_kwargs,
            num_proc=num_proc,
            fingerprint=fingerprint,
        )

    def fit_transform(
        self,
        X,
        *args,
        input_columns: SelectedColumnTypes = None,
        keep_unused_columns: bool = True,
        raise_if_missing: bool = False,
        cache_output: bool = None,
        load_from_cache_file: bool = None,
        batched: bool = True,
        batch_size: int = 1000,
        batch_format: str = None,
        output_format: str = None,
        map_kwargs: dict = {"fn_kwargs": {}},
        num_proc: int = None,
        cache_dir: str = None,
        cache_file_name: str = None,
        fingerprint: str = None,
    ):
        return self.fit(
            X,
            input_columns=input_columns,
            raise_if_missing=raise_if_missing,
            cache_output=cache_output,
            load_from_cache_file=load_from_cache_file,
            batched=batched,
            batch_size=batch_size,
            batch_format=batch_format,
            num_proc=num_proc,
            map_kwargs=map_kwargs,
            cache_dir=cache_dir,
            cache_file_name=cache_file_name,
            fingerprint=fingerprint,
        ).transform(
            X,
            input_columns=input_columns,
            keep_unused_columns=keep_unused_columns,
            raise_if_missing=raise_if_missing,
            cache_output=cache_output,
            cache_dir=cache_dir,
            cache_file_name=cache_file_name,
            load_from_cache_file=load_from_cache_file,
            batched=batched,
            batch_size=batch_size,
            batch_format=batch_format,
            output_format=output_format,
            map_kwargs=map_kwargs,
            num_proc=num_proc,
            fingerprint=fingerprint,
        )

    def _process_fit_input(self, input, **kwargs):
        if (
            self.config.lower_threshold != "auto"
            and self.config.upper_threshold != "auto"
        ):
            kwargs["fn_kwargs"]["fn"] = None
        return super()._process_fit_input(input, **kwargs)

    def _fit_polars(self, X: "pl.DataFrame"):
        iqr = [X.sum_horizontal().quantile(0.25), X.sum_horizontal().quantile(0.75)]
        if self.config.lower_threshold == "auto":
            self.config.lower_threshold = iqr[0] - 1.5 * (iqr[1] - iqr[0])
        if self.config.upper_threshold == "auto":
            self.config.upper_threshold = iqr[1] + 1.5 * (iqr[1] - iqr[0])
        return self

    def _partial_fit_polars(self, X: "pl.DataFrame"):
        iqr = [X.sum_horizontal().quantile(0.25), X.sum_horizontal().quantile(0.75)]
        if self.IQRs is None:
            self.IQRs = [[iqr[0]], [iqr[1]]]
        else:
            self.IQRs[0].append(iqr[0])
            self.IQRs[1].append(iqr[1])
        return self

    def _fit_pandas(self, X: "pd.DataFrame"):
        iqr = [X.sum(axis=1).quantile(0.25), X.sum(axis=1).quantile(0.75)]
        if self.config.lower_threshold == "auto":
            self.config.lower_threshold = iqr[0] - 1.5 * (iqr[1] - iqr[0])
        if self.config.upper_threshold == "auto":
            self.config.upper_threshold = iqr[1] + 1.5 * (iqr[1] - iqr[0])
        return self

    def _partial_fit_pandas(self, X: "pd.DataFrame"):
        iqr = [X.sum(axis=1).quantile(0.25), X.sum(axis=1).quantile(0.75)]
        if self.IQRs is None:
            self.IQRs = [[iqr[0]], [iqr[1]]]
        else:
            self.IQRs[0].append(iqr[0])
            self.IQRs[1].append(iqr[1])
        return self

    def _fit_numpy(self, X: np.ndarray):
        iqr = [np.quantile(X.sum(axis=1), 0.25), np.quantile(X.sum(axis=1), 0.75)]
        if self.config.lower_threshold == "auto":
            self.config.lower_threshold = iqr[0] - 1.5 * (iqr[1] - iqr[0])
        if self.config.upper_threshold == "auto":
            self.config.upper_threshold = iqr[1] + 1.5 * (iqr[1] - iqr[0])
        return self

    def _partial_fit_numpy(self, X: np.ndarray):
        iqr = [np.quantile(X.sum(axis=1), 0.25), np.quantile(X.sum(axis=1), 0.75)]
        if self.IQRs is None:
            self.IQRs = [[iqr[0]], [iqr[1]]]
        else:
            self.IQRs[0].append(iqr[0])
            self.IQRs[1].append(iqr[1])
        return self

    def _pool_fit_any(self, partial_results: List["AbundanceSampleFilter"]):
        IQRs = [[], []]
        for result in partial_results:
            IQRs[0].extend(result.IQRs[0])
            IQRs[1].extend(result.IQRs[1])
        IQRs[0] = np.mean(IQRs[0])
        IQRs[1] = np.mean(IQRs[1])
        if self.config.lower_threshold == "auto":
            self.config.lower_threshold = IQRs[0] - 1.5 * (IQRs[1] - IQRs[0])
        if self.config.upper_threshold == "auto":
            self.config.upper_threshold = IQRs[1] + 1.5 * (IQRs[1] - IQRs[0])
        return self

    def _process_fit_output(self, input, out):
        if out.config.lower_threshold and out.config.upper_threshold:
            logger.info(
                f"Using {out.config.lower_threshold} as lower threshold and "
                f"{out.config.upper_threshold} as upper threshold for filtering samples"
            )
        elif out.config.lower_threshold and out.config.upper_threshold:
            logger.info(
                f"Using {out.config.lower_threshold} as lower threshold for filtering samples"
            )
        elif out.config.upper_threshold:
            logger.info(
                f"Using {out.config.upper_threshold} as upper threshold for filtering samples"
            )
        return super()._process_fit_output(input, out)

    def _process_transform_input(self, input, **kwargs):
        if self.config.lower_threshold and self.config.upper_threshold:
            kwargs["desc"] = (
                f"SampleFiltering out examples with less than {self.config.lower_threshold:.2f} and more than {self.config.upper_threshold:.2f} total abundance"
            )
        elif self.config.lower_threshold:
            kwargs["desc"] = (
                f"SampleFiltering out examples with less than {self.config.lower_threshold:.2f} total abundance"
            )
        elif self.config.upper_threshold:
            kwargs["desc"] = (
                f"SampleFiltering out examples with more than {self.config.upper_threshold:.2f} total abundance"
            )

        return super()._process_transform_input(input, **kwargs)

    def _transform_polars(self, X: "pl.DataFrame"):
        row_sum = X.sum_horizontal()
        if self.config.upper_threshold and self.config.lower_threshold:
            return (row_sum > self.config.lower_threshold) & (
                row_sum < self.config.upper_threshold
            )
        elif self.config.upper_threshold:
            return row_sum < self.config.upper_threshold
        elif self.config.lower_threshold:
            return row_sum > self.config.lower_threshold

    def _transform_pandas(self, X: "pd.DataFrame"):
        row_sum = X.sum(axis=1)
        if self.config.upper_threshold and self.config.lower_threshold:
            return (row_sum > self.config.lower_threshold) & (
                row_sum < self.config.upper_threshold
            )
        elif self.config.upper_threshold:
            return row_sum < self.config.upper_threshold
        elif self.config.lower_threshold:
            return row_sum > self.config.lower_threshold

    def _transform_numpy(self, X: np.ndarray):
        row_sum = X.sum(axis=1)
        if self.config.upper_threshold and self.config.lower_threshold:
            return (row_sum > self.config.lower_threshold) & (
                row_sum < self.config.upper_threshold
            )
        elif self.config.upper_threshold:
            return row_sum < self.config.upper_threshold
        elif self.config.lower_threshold:
            return row_sum > self.config.lower_threshold



from dataclasses import dataclass, field, fields
from typing import TYPE_CHECKING, Callable, Dict, List, Optional, Type, Union

import numpy as np
from biocore import DataHandler
from biocore.utils.import_util import is_lightgbm_available, requires_backends

from biofit.integration.biosets import get_feature
from biofit.processing import SelectedColumnTypes, sync_backup_config
from biofit.utils import logging

from ..models import Model, ModelConfig

logger = logging.get_logger(__name__)

if TYPE_CHECKING:
    from lightgbm import LGBMClassifier, LGBMRegressor
    from lightgbm.sklearn import _LGBM_ScikitCustomObjectiveFunction

if is_lightgbm_available():
    from lightgbm import LGBMClassifier, LGBMRegressor
    from lightgbm.sklearn import _LGBM_ScikitCustomObjectiveFunction


@dataclass
class LightGBMConfig(ModelConfig):
    _fit_input_feature_types: List[Type] = field(
        default_factory=lambda: [
            None,
            get_feature("TARGET_FEATURE_TYPES"),
            None,
            None,
        ],
        init=False,
        repr=False,
    )
    _fit_unused_feature_types: List[Type] = field(
        default_factory=lambda: [
            get_feature("METADATA_FEATURE_TYPES"),
            None,
            None,
            None,
        ],
        init=False,
        repr=False,
    )
    _transform_unused_feature_types: List[Type] = field(
        default_factory=lambda: [get_feature("METADATA_FEATURE_TYPES")],
        init=False,
        repr=False,
    )
    processor_name: str = field(default="lightgbm", init=False, repr=False)
    boosting_type: str = "gbdt"
    num_leaves: int = 31
    max_depth: int = -1
    learning_rate: float = 0.1
    n_estimators: int = 100
    subsample_for_bin: int = 200000
    objective: Optional[Union[str, "_LGBM_ScikitCustomObjectiveFunction"]] = None
    eval_metric: str = None
    class_weight: Optional[Union[Dict, str]] = None
    min_split_gain: float = 0.0
    min_child_weight: float = 1e-3
    min_child_samples: int = 20
    subsample: float = 1.0
    subsample_freq: int = 0
    colsample_bytree: float = 1.0
    reg_alpha: float = 0.0
    reg_lambda: float = 0.0
    random_state: Optional[Union[int, np.random.RandomState]] = 42
    n_jobs: Optional[int] = None
    importance_type: str = "split"
    early_stopping_rounds: int = None
    callbacks: Optional[List[Callable]] = None
    verbosity: int = -1
    additional_kwargs: dict = field(default_factory=dict)

    n_classes: int = None
    task: str = None
    estimator: Union["LGBMClassifier", "LGBMRegressor"] = field(
        default=None, init=False, repr=False
    )

    def __post_init__(self):
        requires_backends(self.__class__, "lightgbm")
        self._fit_process_desc = f"Fitting the LGBM {self.task} model"
        self.predict_process_desc = f"Predicting with the LGBM {self.task} model"
        self.predict_proba_process_desc = (
            f"Predicting probabilities with the LGBM {self.task} model"
        )


class LightGBMModel(Model):
    config_class = LightGBMConfig
    config: LightGBMConfig
    lightgbm: Union["LGBMClassifier", "LGBMRegressor"]

    def __init__(
        self,
        boosting_type: str = "gbdt",
        num_leaves: int = 31,
        max_depth: int = -1,
        learning_rate: float = 0.1,
        n_estimators: int = 100,
        subsample_for_bin: int = 200000,
        objective: Optional[Union[str, "_LGBM_ScikitCustomObjectiveFunction"]] = None,
        eval_metric: str = None,
        class_weight: Optional[Union[Dict, str]] = None,
        min_split_gain: float = 0.0,
        min_child_weight: float = 1e-3,
        min_child_samples: int = 20,
        subsample: float = 1.0,
        subsample_freq: int = 0,
        colsample_bytree: float = 1.0,
        reg_alpha: float = 0.0,
        reg_lambda: float = 0.0,
        random_state: Optional[Union[int, np.random.RandomState]] = None,
        n_jobs: Optional[int] = None,
        importance_type: str = "split",
        callbacks: Optional[List[Callable]] = None,
        verbosity: int = -1,
        n_classes: int = None,
        early_stopping_rounds: int = None,
        task: str = None,
        version: str = None,
        config: Optional[LightGBMConfig] = None,
        **kwargs,
    ):
        m_params = [f.name for f in fields(LightGBMConfig) if f.init]
        biofit_params = {k: v for k, v in kwargs.items() if k in m_params}
        lightgbm_params = {k: v for k, v in kwargs.items() if k not in m_params}
        if "additional_kwargs" in biofit_params:
            lightgbm_params.update(biofit_params.pop("additional_kwargs"))
        super().__init__(
            config=config,
            boosting_type=boosting_type,
            num_leaves=num_leaves,
            max_depth=max_depth,
            learning_rate=learning_rate,
            n_estimators=n_estimators,
            subsample_for_bin=subsample_for_bin,
            objective=objective,
            eval_metric=eval_metric,
            class_weight=class_weight,
            min_split_gain=min_split_gain,
            min_child_weight=min_child_weight,
            min_child_samples=min_child_samples,
            subsample=subsample,
            subsample_freq=subsample_freq,
            colsample_bytree=colsample_bytree,
            reg_alpha=reg_alpha,
            reg_lambda=reg_lambda,
            random_state=random_state,
            n_jobs=n_jobs,
            importance_type=importance_type,
            early_stopping_rounds=early_stopping_rounds,
            callbacks=callbacks,
            verbosity=verbosity,
            n_classes=n_classes,
            task=task,
            version=version,
            additional_kwargs=lightgbm_params,
            **biofit_params,
        )
        if self.config.task is None:
            raise ValueError(
                "Task is not set. Please set the task before setting the parameters."
            )

        if "classification" in self.config.task:
            self.lightgbm = LGBMClassifier(
                boosting_type=self.config.boosting_type,
                num_leaves=self.config.num_leaves,
                max_depth=self.config.max_depth if self.config.max_depth > 0 else -1,
                learning_rate=self.config.learning_rate,
                n_estimators=self.config.n_estimators,
                subsample_for_bin=self.config.subsample_for_bin,
                objective=self.config.objective,
                class_weight=self.config.class_weight
                if self.config.class_weight != "None"
                else None,
                min_split_gain=self.config.min_split_gain,
                min_child_weight=self.config.min_child_weight,
                min_child_samples=self.config.min_child_samples,
                subsample=self.config.subsample,
                subsample_freq=self.config.subsample_freq,
                colsample_bytree=self.config.colsample_bytree,
                reg_alpha=self.config.reg_alpha,
                reg_lambda=self.config.reg_lambda,
                random_state=self.config.random_state,
                n_jobs=self.config.n_jobs,
                importance_type=self.config.importance_type,
                verbosity=self.config.verbosity,
                **self.config.additional_kwargs,
            )
        elif "regression" in self.config.task:
            self.lightgbm = LGBMRegressor(
                boosting_type=self.config.boosting_type,
                num_leaves=self.config.num_leaves,
                max_depth=self.config.max_depth if self.config.max_depth > 0 else -1,
                learning_rate=self.config.learning_rate,
                n_estimators=self.config.n_estimators,
                subsample_for_bin=self.config.subsample_for_bin,
                objective=self.config.objective,
                min_split_gain=self.config.min_split_gain,
                min_child_weight=self.config.min_child_weight,
                min_child_samples=self.config.min_child_samples,
                subsample=self.config.subsample,
                subsample_freq=self.config.subsample_freq,
                colsample_bytree=self.config.colsample_bytree,
                reg_alpha=self.config.reg_alpha,
                reg_lambda=self.config.reg_lambda,
                random_state=self.config.random_state,
                n_jobs=self.config.n_jobs,
                importance_type=self.config.importance_type,
                verbosity=self.config.verbosity,
                **self.config.additional_kwargs,
            )
        else:
            raise ValueError(f"Invalid task: {self.config.task}")

    @sync_backup_config
    def set_params(self, **params):
        self.config = self.config.replace_defaults(**params)
        if self.config.task is None:
            raise ValueError(
                "Task is not set. Please set the task before setting the parameters."
            )

        if "classification" in self.config.task:
            self.lightgbm = LGBMClassifier(
                boosting_type=self.config.boosting_type,
                num_leaves=self.config.num_leaves,
                max_depth=self.config.max_depth if self.config.max_depth > 0 else -1,
                learning_rate=self.config.learning_rate,
                n_estimators=self.config.n_estimators,
                subsample_for_bin=self.config.subsample_for_bin,
                objective=self.config.objective,
                class_weight=self.config.class_weight
                if self.config.class_weight != "None"
                else None,
                min_split_gain=self.config.min_split_gain,
                min_child_weight=self.config.min_child_weight,
                min_child_samples=self.config.min_child_samples,
                subsample=self.config.subsample,
                subsample_freq=self.config.subsample_freq,
                colsample_bytree=self.config.colsample_bytree,
                reg_alpha=self.config.reg_alpha,
                reg_lambda=self.config.reg_lambda,
                random_state=self.config.random_state,
                n_jobs=self.config.n_jobs,
                importance_type=self.config.importance_type,
                verbosity=self.config.verbosity,
            )
        elif "regression" in self.config.task:
            self.lightgbm = LGBMRegressor(
                boosting_type=self.config.boosting_type,
                num_leaves=self.config.num_leaves,
                max_depth=self.config.max_depth if self.config.max_depth > 0 else -1,
                learning_rate=self.config.learning_rate,
                n_estimators=self.config.n_estimators,
                subsample_for_bin=self.config.subsample_for_bin,
                objective=self.config.objective,
                min_split_gain=self.config.min_split_gain,
                min_child_weight=self.config.min_child_weight,
                min_child_samples=self.config.min_child_samples,
                subsample=self.config.subsample,
                subsample_freq=self.config.subsample_freq,
                colsample_bytree=self.config.colsample_bytree,
                reg_alpha=self.config.reg_alpha,
                reg_lambda=self.config.reg_lambda,
                random_state=self.config.random_state,
                n_jobs=self.config.n_jobs,
                importance_type=self.config.importance_type,
                verbosity=self.config.verbosity,
            )
        else:
            raise ValueError(f"Invalid task: {self.config.task}")
        return self

    def set_objective(self, task: str):
        if task == "regression":
            if self.config.eval_metric == "rmse":
                self.config.objective = "regression"
            elif self.config.eval_metric == "mae":
                self.config.objective = "regression_l1"
        elif task == "binary_classification":
            self.config.objective = "binary"
            if self.config.eval_metric == "accuracy":
                self.config.eval_metric = "binary_error"
        elif task == "multiclass_classification":
            self.config.objective = "multiclass"
        self.lightgbm = self.lightgbm.set_params(objective=self.config.objective)
        return self

    @property
    def feature_importances_(self):
        return self.config.estimator.feature_importances_

    @property
    def feature_names_in_(self):
        return self.config.feature_names_in_

    def fit(
        self,
        X,
        y=None,
        eval_set=None,
        input_columns: SelectedColumnTypes = None,
        target_column: SelectedColumnTypes = None,
        eval_input_columns: SelectedColumnTypes = None,
        eval_target_column: SelectedColumnTypes = None,
        raise_if_missing: bool = None,
        cache_output: bool = None,
        load_from_cache_file: bool = None,
        batched: bool = True,
        batch_size: int = 1000,
        batch_format: str = None,
        num_proc: int = None,
        map_kwargs: dict = {"fn_kwargs": {}},
        cache_dir: str = None,
        cache_file_name: str = None,
        fingerprint: str = None,
        early_stopping_rounds: int = None,
    ) -> "LightGBMModel":
        self.config._input_columns = self._set_input_columns_and_arity(
            input_columns, target_column, eval_input_columns, eval_target_column
        )
        if eval_set is not None:
            if isinstance(eval_set, tuple):
                extras = eval_set
            elif (
                isinstance(eval_set, list)
                and len(eval_set) == 1
                and isinstance(eval_set[0], tuple)
            ):
                extras = eval_set[0]
            else:
                raise ValueError(
                    "eval_set must be a tuple or a list containing a single tuple"
                )
        else:
            extras = (None, None)

        return self._process_fit(
            X,
            y,
            *extras,
            raise_if_missing=raise_if_missing,
            cache_output=cache_output,
            load_from_cache_file=load_from_cache_file,
            batched=batched,
            batch_size=batch_size,
            batch_format=batch_format,
            num_proc=num_proc,
            map_kwargs=map_kwargs,
            cache_dir=cache_dir,
            cache_file_name=cache_file_name,
            fingerprint=fingerprint,
        )

    def predict(
        self,
        X,
        input_columns: SelectedColumnTypes = None,
        keep_unused_columns: bool = None,
        raise_if_missing: bool = None,
        cache_output: bool = None,
        cache_dir: str = None,
        cache_file_name: str = None,
        load_from_cache_file: bool = None,
        batched: bool = None,
        batch_size: int = None,
        batch_format: str = None,
        output_format: str = None,
        map_kwargs: dict = None,
        num_proc: int = None,
        fingerprint: str = None,
    ):
        self._method_prefix = "_predict"
        self._input_columns = self._set_input_columns_and_arity(input_columns)
        return self._process_transform(
            X,
            keep_unused_columns=keep_unused_columns,
            raise_if_missing=raise_if_missing,
            cache_output=cache_output,
            cache_dir=cache_dir,
            cache_file_name=cache_file_name,
            load_from_cache_file=load_from_cache_file,
            batched=batched,
            batch_size=batch_size,
            batch_format=batch_format,
            output_format=output_format,
            map_kwargs=map_kwargs,
            num_proc=num_proc,
            fingerprint=fingerprint,
        )

    def predict_proba(
        self,
        X,
        input_columns: SelectedColumnTypes = None,
        keep_unused_columns: bool = None,
        raise_if_missing: bool = None,
        cache_output: bool = None,
        cache_dir: str = None,
        cache_file_name: str = None,
        load_from_cache_file: bool = None,
        batched: bool = None,
        batch_size: int = None,
        batch_format: str = None,
        output_format: str = None,
        map_kwargs: dict = None,
        num_proc: int = None,
        fingerprint: str = None,
    ):
        self._method_prefix = "_predict_proba"
        self._input_columns = self._set_input_columns_and_arity(input_columns)
        return self._process_transform(
            X,
            keep_unused_columns=keep_unused_columns,
            raise_if_missing=raise_if_missing,
            cache_output=cache_output,
            cache_dir=cache_dir,
            cache_file_name=cache_file_name,
            load_from_cache_file=load_from_cache_file,
            batched=batched,
            batch_size=batch_size,
            batch_format=batch_format,
            output_format=output_format,
            map_kwargs=map_kwargs,
            num_proc=num_proc,
            fingerprint=fingerprint,
        )

    def _fit_sklearn(self, X, y, eval_x=None, eval_y=None):
        callbacks = self.config.callbacks
        if self.config.early_stopping_rounds:
            callbacks = callbacks or []
            from lightgbm.callback import early_stopping

            callbacks.append(early_stopping(self.config.early_stopping_rounds))

        if eval_x is not None and eval_y is not None:
            self.config.estimator = self.lightgbm.fit(
                X,
                y,
                eval_set=[(eval_x, eval_y)],
                eval_metric=self.config.eval_metric,
                callbacks=callbacks,
            )
        else:
            self.config.estimator = self.lightgbm.fit(
                X, y, eval_metric=self.config.eval_metric, callbacks=callbacks
            )
        if self.config.early_stopping_rounds:
            self.config.n_estimators = self.config.estimator.best_iteration_
            self.config.estimator.set_params(n_estimators=self.config.n_estimators)
            self.config.early_stopping_rounds = None
        return self

    def _predict_sklearn(self, X):
        return self.config.estimator.predict(X)

    def _predict_proba_sklearn(self, X):
        return self.config.estimator.predict_proba(X)

    def _get_features_out(
        self, X, selected_indices=None, unselected_indices=None, **kwargs
    ):
        if self._method_prefix == "_predict_proba":
            self.config._n_features_out = self.config.n_classes
            self.output_dtype = "float64"
        else:
            self.config._n_features_out = 1
        return super()._get_features_out(
            X,
            selected_indices=selected_indices,
            unselected_indices=unselected_indices,
            one_to_one_features=False,
            n_features_out=self.config._n_features_out,
            keep_unused_columns=False,
        )

    @staticmethod
    def suggest_params(data):
        params = {
            "verbosity": -1,
            "n_estimators": ("suggest_int", [50, 1000], {"step": 50}),
            "max_depth": ("suggest_int", [0, 11], {}),
            "learning_rate": ("suggest_float", [1e-3, 0.1], {"log": True}),
            "num_leaves": ("suggest_int", [31, 255], {}),
            "subsample": ("suggest_float", [0.5, 1.0], {"step": 0.1}),
            "reg_alpha": ("suggest_float", [1e-9, 1.0], {"log": True}),
            "reg_lambda": ("suggest_float", [1e-9, 1.0], {"log": True}),
            "min_child_samples": ("suggest_int", [10, 100], {}),
            "min_split_gain": ("suggest_float", [0.0, 1.0], {}),
            "num_threads": -1,
        }

        if data is not None:
            data_dim = DataHandler.get_shape(data)

        if data_dim[1] > data_dim[0]:
            msg = "The number of features is greater than the number of samples"
            max_feat_frac = max(0.5, data_dim[0] / data_dim[1])
            max_feat_frac = round(max_feat_frac, 1)
            params["colsample_bytree"] = (
                "suggest_float",
                [0.1, max_feat_frac],
                {"step": 0.1},
            )
            if data_dim[1] > 1000:
                msg += " and exceeds 1000"
                params["n_estimators"] = ("suggest_int", [50, 500], {"step": 50})
                params["max_depth"] = ("suggest_int", [3, 8], {})
                params["num_leaves"] = ("suggest_int", [31, 63], {})
                params["min_child_samples"] = ("suggest_int", [10, 50], {})
                params["max_bin"] = ("suggest_int", [63, 127], {})
            msg += ". Adjusting the hyperparameters accordingly for faster training. "
            logger.warning(msg)

        return params

    @staticmethod
    def suggest_first_trial():
        return {
            "n_estimators": 100,
            "max_depth": 8,
            "learning_rate": 0.1,
            "num_leaves": 31,
            "subsample": 1.0,
            "reg_alpha": 1e-9,
            "reg_lambda": 1e-9,
            "min_child_samples": 20,
            "colsample_bytree": 0.9,
            "min_split_gain": 0.0,
        }


class LightGBMForClassification(LightGBMModel):
    def __init__(self, **kwargs):
        kwargs["task"] = kwargs.get("task", "classification")
        super().__init__(**kwargs)

    @property
    def classes_(self):
        return self.config.estimator.classes_

    @sync_backup_config
    def set_params(self, **params):
        self.config.task = self.config.task or "classification"
        return super().set_params(**params)

    def _process_fit_input(self, X, **kwargs):
        X, kwargs = super()._process_fit_input(X, **kwargs)
        self.set_objective(self.config.task)
        return X, kwargs

    @staticmethod
    def suggest_params(data):
        params = LightGBMModel.suggest_params(data)
        params["class_weight"] = ("suggest_categorical", [["balanced", "None"]], {})
        return params

    @staticmethod
    def suggest_first_trial():
        return {
            **LightGBMModel.suggest_first_trial(),
            "class_weight": "balanced",
        }


class LightGBMForRegression(LightGBMModel):
    @sync_backup_config
    def set_params(self, **params):
        self.config.task = self.config.task or "regression"
        return super().set_params(**params)



# ruff: noqa
from .row_abundance import (
    AbundanceSampleFilterConfigForOTU,
    AbundanceSampleFilterConfig,
    AbundanceSampleFilter,
)
from .plot_row_abundance import (
    AbundanceSampleFilterPlotterConfigForOTU,
    AbundanceSampleFilterPlotterConfig,
    AbundanceSampleFilterPlotter,
)



from dataclasses import dataclass, field
from typing import TYPE_CHECKING, List, Type

import numpy as np

from biofit.integration.biosets import get_feature
from biofit.processing import SelectedColumnTypes
from biofit.utils import logging

from ..stat import Stat, StatConfig

if TYPE_CHECKING:
    import polars as pl

logger = logging.get_logger(__name__)


@dataclass
class RowMeanStatConfig(StatConfig):
    _transform_unused_feature_types: List[Type] = field(
        default_factory=lambda: [get_feature("METADATA_FEATURE_TYPES")],
        init=False,
        repr=False,
    )
    # process description
    _transform_process_desc: str = field(
        default="Calculating row means", init=False, repr=False
    )
    processor_name: str = field(default="row_mean", init=False, repr=False)
    _n_features_out: int = field(default=1, init=False, repr=False)


@dataclass
class RowMeanStatConfigForOTU(RowMeanStatConfig):
    """Computes the sum of each row in the OTUAbundance feature.
    This class is the same as TotalOTUAbundanceStat.
    It is provided for autostat
    """

    # dataset specific attributes
    _transform_input_feature_types: List[Type] = field(
        default_factory=lambda: [get_feature("Abundance")], init=False, repr=False
    )
    dataset_name: str = field(default="otu", init=False, repr=False)


class RowMeanStat(Stat):
    # config attributes
    _config_class = RowMeanStatConfig
    config: RowMeanStatConfig
    output_dtype = "float64"

    def fit(
        self,
        X,
        input_columns: SelectedColumnTypes = None,
        raise_if_missing: bool = None,
        cache_output: bool = None,
        load_from_cache_file: bool = None,
        batched: bool = True,
        batch_size: int = 1000,
        batch_format: str = None,
        num_proc: int = None,
        map_kwargs: dict = {"fn_kwargs": {}},
        cache_dir: str = None,
        cache_file_name: str = None,
        fingerprint: str = None,
    ) -> "RowMeanStat":
        self.config._input_columns = self._set_input_columns_and_arity(input_columns)
        return self._process_fit(
            X,
            raise_if_missing=raise_if_missing,
            cache_output=cache_output,
            load_from_cache_file=load_from_cache_file,
            batched=batched,
            batch_size=batch_size,
            batch_format=batch_format,
            num_proc=num_proc,
            map_kwargs=map_kwargs,
            cache_dir=cache_dir,
            cache_file_name=cache_file_name,
            fingerprint=fingerprint,
        )

    def transform(
        self,
        X,
        input_columns: SelectedColumnTypes = None,
        keep_unused_columns: bool = None,
        raise_if_missing: bool = None,
        cache_output: bool = None,
        cache_dir: str = None,
        cache_file_name: str = None,
        load_from_cache_file: bool = None,
        batched: bool = None,
        batch_size: int = None,
        batch_format: str = None,
        output_format: str = None,
        map_kwargs: dict = None,
        num_proc: int = None,
        fingerprint: str = None,
    ):
        self._input_columns = self._set_input_columns_and_arity(input_columns)
        return self._process_transform(
            X,
            keep_unused_columns=keep_unused_columns,
            raise_if_missing=raise_if_missing,
            cache_output=cache_output,
            cache_dir=cache_dir,
            cache_file_name=cache_file_name,
            load_from_cache_file=load_from_cache_file,
            batched=batched,
            batch_size=batch_size,
            batch_format=batch_format,
            output_format=output_format,
            map_kwargs=map_kwargs,
            num_proc=num_proc,
            fingerprint=fingerprint,
        )

    def fit_transform(
        self,
        X,
        *args,
        input_columns: SelectedColumnTypes = None,
        keep_unused_columns: bool = True,
        raise_if_missing: bool = False,
        cache_output: bool = None,
        load_from_cache_file: bool = None,
        batched: bool = True,
        batch_size: int = 1000,
        batch_format: str = None,
        output_format: str = None,
        map_kwargs: dict = {"fn_kwargs": {}},
        num_proc: int = None,
        cache_dir: str = None,
        cache_file_name: str = None,
        fingerprint: str = None,
    ):
        return self.fit(
            X,
            input_columns=input_columns,
            raise_if_missing=raise_if_missing,
            cache_output=cache_output,
            load_from_cache_file=load_from_cache_file,
            batched=batched,
            batch_size=batch_size,
            batch_format=batch_format,
            num_proc=num_proc,
            map_kwargs=map_kwargs,
            cache_dir=cache_dir,
            cache_file_name=cache_file_name,
            fingerprint=fingerprint,
        ).transform(
            X,
            input_columns=input_columns,
            keep_unused_columns=keep_unused_columns,
            raise_if_missing=raise_if_missing,
            cache_output=cache_output,
            cache_dir=cache_dir,
            cache_file_name=cache_file_name,
            load_from_cache_file=load_from_cache_file,
            batched=batched,
            batch_size=batch_size,
            batch_format=batch_format,
            output_format=output_format,
            map_kwargs=map_kwargs,
            num_proc=num_proc,
            fingerprint=fingerprint,
        )

    def _transform_numpy(self, X: np.ndarray, y=None):
        if len(X.shape) == 1:
            return np.mean(X)[:, None]
        return np.mean(X, axis=1)[:, None]

    def _transform_polars(self, X: "pl.DataFrame", y=None):
        return X.mean_horizontal().to_frame()



import copy
import os
from pathlib import Path
from typing import TYPE_CHECKING, Callable, List, Union

import numpy as np
import pandas as pd
from biocore import DataHandler
from sklearn.base import BaseEstimator
from sklearn.model_selection import (
    BaseCrossValidator,
)
from sklearn.pipeline import Pipeline

from biofit.auto.processing_auto import ProcessorPipeline
from biofit.processing import BaseProcessor
from biofit.train_eval_utils import (
    _flatten_pipeline,
    _get_data,
    split,
)
from biofit.utils import (
    enable_full_determinism,
    logging,
)

if TYPE_CHECKING:
    import polars as pl
    from datasets import Dataset

logger = logging.get_logger(__name__)


def preprocess(
    preprocessor,
    x_train,
    y_train=None,
    x_valid=None,
    y_valid=None,
    cache_dir=None,
    transform_only=False,
    raise_error=False,
):
    try:
        if isinstance(preprocessor, (Pipeline, ProcessorPipeline)):
            for proc in preprocessor.steps:
                p = proc[-1] if isinstance(proc, tuple) else proc
                if isinstance(p, BaseProcessor):
                    extra_kwargs = {
                        "cache_dir": cache_dir,
                        # "load_from_cache_file": False,
                    }
                    # p.config.enable_caching = False
                else:
                    extra_kwargs = {}
                if transform_only:
                    x_train = p.transform(x_train, **extra_kwargs)
                else:
                    x_train = p.fit_transform(x_train, **extra_kwargs)
                if x_valid is not None:
                    x_valid = p.transform(x_valid, **extra_kwargs)
        else:
            if isinstance(preprocessor, BaseProcessor):
                extra_kwargs = {
                    "cache_dir": cache_dir,
                    # "load_from_cache_file": False,
                }
                # p.config.enable_caching = False
            else:
                extra_kwargs = {}
            if transform_only:
                x_train = preprocessor.transform(x_train, **extra_kwargs)
            else:
                x_train = preprocessor.fit_transform(x_train, **extra_kwargs)
            if x_valid is not None:
                x_valid = preprocessor.transform(x_valid, **extra_kwargs)
    except ValueError:
        if raise_error:
            raise
        else:
            logger.info("Preprocessing failed, using nan_to_num")
            _train_cols = x_train.columns.tolist()
            x_train = np.nan_to_num(x_train)
            # convert back to dataframe
            x_train = pd.DataFrame(x_train, columns=_train_cols)
            if x_valid is not None:
                _valid_cols = x_valid.columns.tolist()
                x_valid = np.nan_to_num(x_valid)
                x_valid = pd.DataFrame(x_valid, columns=_valid_cols)
            return preprocess(
                preprocessor,
                x_train,
                y_train,
                x_valid,
                y_valid,
                cache_dir,
                transform_only,
                True,
            )

    return x_train, y_train, x_valid, y_valid


def train(
    model: Union[BaseEstimator, BaseProcessor],
    data: Union[pd.DataFrame, "pl.DataFrame", "Dataset"],
    target: Union[
        pd.Series, "pl.Series", pd.DataFrame, "pl.DataFrame", "Dataset"
    ] = None,
    valid_data: Union[pd.DataFrame, "pl.DataFrame", "Dataset"] = None,
    valid_target: Union[
        pd.Series, "pl.Series", pd.DataFrame, "pl.DataFrame", "Dataset"
    ] = None,
    groups: Union[pd.Series, "pl.Series"] = None,
    input_columns: Union[List[str], str] = "auto",
    target_columns: Union[List[str], str] = "auto",
    group_name: str = None,
    preprocessor: Union[BaseEstimator, BaseProcessor] = None,
    eval_metric: Union[str, Callable] = None,
    task: str = None,
    cv: BaseCrossValidator = None,
    random_state: Union[List[int], int] = 42,
    save_indices: bool = False,
    output_dir: str = None,
    cache_dir: str = None,
):
    """
    Train a model or processor on the provided data, with optional preprocessing,
    validation data, and cross-validation.

    This function supports training models with various data formats (pandas DataFrame,
    polars DataFrame, Dataset), optional preprocessing pipelines, cross-validation
    strategies, and random seeds for reproducibility.

    Parameters:
        model (Union[BaseEstimator, BaseProcessor]):
            The model or processor to train. This can be a scikit-learn estimator, a
            pipeline, or a custom estimator that follows scikit-learn's API.
        data (Union[pd.DataFrame, pl.DataFrame, "Dataset"]):
            The training data. It can be a pandas DataFrame, polars DataFrame, or a
            Dataset object.
        target (Union[pd.Series, pl.Series, pd.DataFrame, pl.DataFrame, "Dataset"], optional):
            The target values for supervised learning. If None, the target columns must
            be specified in `data` using `target_columns`.
        valid_data (Union[pd.DataFrame, pl.DataFrame, "Dataset"], optional):
            Optional validation data. If not provided, validation can be done using
            cross-validation.
        valid_target (Union[pd.Series, pl.Series, pd.DataFrame, pl.DataFrame, "Dataset"], optional):
            Target values for the validation data.
        groups (Union[pd.Series, pl.Series], optional):
            Group labels for the samples used while splitting the dataset into
            train/test set.
        input_columns (Union[List[str], str], optional):
            Names of the input feature columns. If 'auto', the function will attempt to
            infer the input columns from the data. Defaults to "auto".
        target_columns (Union[List[str], str], optional):
            Names of the target columns in `data`. If 'auto', the function will attempt
            to infer the target columns. Defaults to "auto".
        group_name (str, optional):
            Name of the group column in `data` if groups are specified within the data.
        preprocessor (Union[BaseEstimator, BaseProcessor], optional):
            Preprocessing pipeline or estimator to apply before training the model.
        eval_metric (Union[str, Callable], optional):
            Evaluation metric or a callable function to evaluate the model's
            performance.
        task (str, optional):
            Type of task to perform (e.g., 'classification', 'regression',
            'multilabel_classification', 'multi_regression').
        cv (type of BaseCrossValidator, optional):
            Cross-validation splitting strategy. If provided, the model will be trained
            using cross-validation.
        random_state (Union[List[int], int], optional):
            Random seed(s) for reproducibility. Can be a single integer or a list of
            integers. Defaults to 42.
        save_indices (bool, optional):
            Whether to save the indices of the train and validation splits. Defaults to
            False.
        output_dir (str, optional):
            Directory where outputs will be saved. Defaults to None.
        cache_dir (str, optional):
            Directory where cache files will be stored. Defaults to None.

    Returns:
        Trained model or pipeline. The return type depends on the inputs:

        - If `random_state` is an integer and `cv` is None, returns a single trained model or pipeline.
        - If `random_state` is a list of integers, returns a list of trained models or pipelines, one for each random seed.
        - If `cv` is specified, returns a list of trained models or pipelines, one for each cross-validation fold.

    Examples:
        1. Basic usage with pandas DataFrame:

        ```python
        from biofit.trainer import train
        from biofit.models import LogisticRegressionForClassification
        import pandas as pd

        # Create sample data
        data = pd.DataFrame({
            'age': [25, 32, 47, 51, 62],
            'salary': [50000, 60000, 70000, 80000, 90000],
            'purchased': [0, 1, 0, 1, 0]
        })

        # Train the model
        model = train(
            model=LogisticRegressionForClassification(),
            data=data,
            target_columns='purchased',
            input_columns=['age', 'salary']
        )
        ```

        2. Using validation data:

        ```python
        from biofit.train import train
        from biofit.models import RandomForestForClassification
        import pandas as pd

        # Training data
        train_data = pd.DataFrame({
            'feature': [1,2,3,4,5],
            'target': [2,4,6,8,10]
        })

        # Validation data
        valid_data = pd.DataFrame({
            'feature': [6,7],
            'target': [12,14]
        })

        # Train the model with validation data
        model = train(
            model=RandomForestForClassification(),
            data=train_data,
            valid_data=valid_data,
            target_columns='target',
            input_columns='feature'
        )
        ```

        3. Cross-validation with random seed list:

        ```python
        from biofit.train import train
        from biofit.models import RandomForestForClassifier
        from sklearn.model_selection import StratifiedKFold
        import pandas as pd

        # Sample data
        data = pd.DataFrame({
            'feature1': [1,2,3,4,5,6],
            'feature2': [6,5,4,3,2,1],
            'target': [0,1,0,1,0,1]
        })

        cv = StratifiedKFold(n_splits=3)
        random_seeds = [42, 7, 21]

        # Train the model with cross-validation and multiple random seeds
        models = train(
            model=RandomForestForClassifier(),
            data=data,
            target_columns='target',
            input_columns=['feature1', 'feature2'],
            cv=cv,
            random_state=random_seeds
        )

        # models is a list of models trained with different random seeds
        ```

        4. Using a preprocessor pipeline:

        ```python
        from biofit.train import train
        from biofit.auto.processing_auto import ProcessorPipeline
        from biofit.models import LogisticRegressionForClassification
        from sklearn.compose import ColumnTransformer
        import pandas as pd

        # Sample data with categorical feature
        data = pd.DataFrame({
            'age': [25, 32, 47, 51, 62],
            'gender': ['M', 'F', 'M', 'F', 'M'],
            'purchased': [0, 1, 0, 1, 0]
        })

        # Define preprocessing
        numeric_features = ['age']
        numeric_transformer = StandardScaler()

        categorical_features = ['gender']
        categorical_transformer = OneHotEncoder()

        preprocessor = ColumnTransformer(
            transformers=[
                ('num', numeric_transformer, numeric_features),
                ('cat', categorical_transformer, categorical_features)
            ])

        # Create a pipeline that first preprocesses the data, then applies the model
        pipeline = ProcessorPipeline(steps=[
            ('preprocessor', preprocessor),
            ('classifier', LogisticRegressionForClassification())
        ])

        # Train the model with preprocessor
        model = train(
            model=pipeline,
            data=data,
            target_columns='purchased',
            input_columns=['age', 'gender']
        )
        ```

        5. Training with a Dataset object:

        ```python
        from biofit.train import train
        from datasets import load_dataset

        # Load a dataset from the datasets library
        dataset = load_dataset('csv', data_files='data.csv')

        # Train the model
        model = train(
            model=SomeModel(),
            data=dataset['train'],
            target_columns='label',
            input_columns='text'
        )
        ```

    Notes:
        - The `train` function can handle various data formats and will attempt to
            process the data accordingly.
        - If `input_columns` or `target_columns` are set to 'auto', the function will
            try to infer the columns automatically based on the data.
        - The `preprocessor` can be any transformer that follows the scikit-learn API,
            including pipelines.
        - When using cross-validation (`cv` is not None), the function will train models
            for each fold and return a list of models.
        - If multiple random seeds are provided (`random_state` is a list), the function
            will train multiple models for feature importance analysis, but only return
            the models trained with the first random seed.
    """
    x_train, y_train, x_valid, y_valid, groups, _, input_columns, target_columns = (
        _get_data(
            data=data,
            target=target,
            valid_data=valid_data,
            valid_target=valid_target,
            groups=groups,
            group_name=group_name,
            input_columns=input_columns,
            target_columns=target_columns,
            format="pandas",
            target_required=True,
        )
    )
    if cache_dir is None and output_dir is not None:
        cache_dir = (Path(output_dir) / ".cache").resolve().as_posix()

    def fit(
        x_train,
        y_train,
        x_valid,
        y_valid,
        model,
        preprocessor,
        random_state=None,
        cache_dir=None,
    ):
        if isinstance(model, Pipeline):
            if preprocessor is None and len(model.steps) > 1:
                preprocessor = _flatten_pipeline(
                    [p[-1] if isinstance(p, tuple) else p for p in model.steps[:-1]]
                )
                preprocessor = Pipeline(preprocessor)
            model = (
                model.steps[-1][1]
                if isinstance(model.steps[-1], tuple)
                else model.steps[-1]
            )
        if preprocessor:
            x_train, y_train, x_valid, y_valid = preprocess(
                preprocessor, x_train, y_train, x_valid, y_valid, cache_dir
            )

        m_params = (
            model.get_params() if hasattr(model, "get_params") else model.__dict__
        )
        if hasattr(model, "set_params"):
            if "random_state" in m_params:
                model.set_params(random_state=random_state)
            elif "seed" in m_params:
                model.set_params(seed=random_state)
        else:
            if "random_state" in m_params:
                model.random_state = random_state
            elif "seed" in m_params:
                model.seed = random_state
        models = []
        if task in ("multilabel_classification", "multi_regression"):
            # also multi_column_regression
            models = [model] * y_train.shape[1]
            for idx, _m in enumerate(models):
                if _m.__module__.startswith("xgboost"):
                    _m.fit(
                        x_train,
                        y_train[:, idx],
                        model__eval_set=[(x_valid, y_valid[:, idx])]
                        if x_valid is not None
                        else None,
                        model__verbose=False,
                    )
                elif _m.__module__.startswith("biofit"):
                    old_val = _m.config.enable_caching
                    _m.config.enable_caching = False
                    if (
                        "early_stopping_rounds" in m_params
                        and m_params["early_stopping_rounds"] is not None
                        and m_params["early_stopping_rounds"] > 0
                    ):
                        _m.fit(
                            x_train,
                            y_train[:, idx],
                            eval_set=[(x_valid, y_valid[:, idx])]
                            if x_valid is not None
                            else None,
                            # load_from_cache_file=False,
                        )
                    else:
                        _m.fit(
                            x_train,
                            y_train[:, idx],
                            load_from_cache_file=False,
                        )
                    _m.config.enable_caching = old_val
                else:
                    _m.fit(
                        x_train,
                        y_train[:, idx],
                    )

        else:
            if model.__module__.startswith("xgboost"):
                model.fit(
                    x_train,
                    y_train,
                    model__eval_set=[(x_valid, y_valid)],
                    model__verbose=False,
                )
            elif model.__module__.startswith("biofit"):
                old_val = model.config.enable_caching
                model.config.enable_caching = False
                model.fit(
                    x_train,
                    y_train,
                    load_from_cache_file=False,
                )
                model.config.enable_caching = old_val
            else:
                model.fit(x_train, y_train)

            models = [model]
        return models, preprocessor

    count = 0
    count += 1

    def training_loop(random_state):
        enable_full_determinism(random_state)
        if cv is None:
            _m = copy.deepcopy(model)
            _p = copy.deepcopy(preprocessor)

            fitted_model_, preprocessor_ = fit(
                x_train=x_train,
                y_train=y_train,
                x_valid=x_valid,
                y_valid=y_valid,
                model=_m,
                preprocessor=_p,
                random_state=random_state,
                cache_dir=cache_dir,
            )

            if isinstance(fitted_model_, list) and len(fitted_model_) == 1:
                fitted_model_ = fitted_model_[0]
            return fitted_model_, preprocessor_
        else:
            fitted_models = []
            fitted_preprocessors = []
            if hasattr(cv, "__dict__") and "random_state" in cv.__dict__:
                cv.random_state = random_state
            for fold, (train_idx, valid_idx) in enumerate(
                split(cv, X=x_train, y=y_train, groups=groups)
            ):
                _cache_dir = os.path.join(cache_dir, f"fold_{fold}")
                _m = copy.deepcopy(model)
                _p = copy.deepcopy(preprocessor)

                m_params = _m.get_params()
                if "random_state" in m_params:
                    _m.set_params(random_state=random_state)
                elif "seed" in m_params:
                    _m.set_params(seed=random_state)
                xtrain_fold, xvalid_fold = (
                    DataHandler.select_rows(x_train, train_idx),
                    DataHandler.select_rows(x_train, valid_idx),
                )
                ytrain_fold, yvalid_fold = (
                    DataHandler.select_rows(y_train, train_idx),
                    DataHandler.select_rows(y_train, valid_idx),
                )
                fitted_model_fold, fitted_preprocessor_fold = fit(
                    xtrain_fold,
                    ytrain_fold,
                    xvalid_fold,
                    yvalid_fold,
                    _m,
                    _p,
                    random_state=random_state,
                    cache_dir=_cache_dir,
                )
                if isinstance(fitted_model_fold, list) and len(fitted_model_fold) == 1:
                    fitted_model_fold = fitted_model_fold[0]
                fitted_models.append(fitted_model_fold)
                fitted_preprocessors.append(fitted_preprocessor_fold)

        return fitted_models, fitted_preprocessors

    def create_pipeline(model, preprocessor):
        pipeline = []
        if isinstance(model, list):
            for m, p in zip(model, preprocessor):
                if isinstance(p, (Pipeline, ProcessorPipeline)):
                    p = ProcessorPipeline(_flatten_pipeline(p))
                pipeline.append(
                    Pipeline([("preprocessor", p), ("model", m)]) if p else m
                )
        else:
            if isinstance(preprocessor, (Pipeline, ProcessorPipeline)):
                preprocessor = ProcessorPipeline(_flatten_pipeline(preprocessor))
            pipeline = (
                Pipeline([("preprocessor", preprocessor), ("model", model)])
                if preprocessor
                else model
            )
        return pipeline

    if isinstance(random_state, list):
        pipeline_list = []
        for rs in random_state:
            m, p = training_loop(rs)
            pipeline_list.append(create_pipeline(m, p))

        return pipeline_list
    else:
        fitted_model, fitted_preprocessor = training_loop(random_state)
        return create_pipeline(fitted_model, fitted_preprocessor)



from dataclasses import dataclass, field
from pathlib import Path
from typing import TYPE_CHECKING, List, Optional, Type

import numpy as np
import pandas as pd
from biocore import DataHandler

from biofit.integration.biosets import get_feature
from biofit.integration.R.r_caller import RCaller
from biofit.processing import SelectedColumnTypes
from biofit.utils.types import Unset
from biofit.visualization.plotting import (
    BasePlotter,
    PlotterConfig,
)

if TYPE_CHECKING:
    from biosets import Dataset


@dataclass
class FeatureImportancePlotterConfig(PlotterConfig):
    _fit_input_feature_types: List[Type] = field(
        default_factory=lambda: [
            None,
            None,
            get_feature("TARGET_FEATURE_TYPES"),
            get_feature("METADATA_FEATURE_TYPES_NOT_TARGET"),
        ],
        init=False,
        repr=False,
    )
    _transform_input_feature_types: List[Type] = field(
        default_factory=lambda: [
            None,
            None,
            get_feature("TARGET_FEATURE_TYPES"),
            get_feature("METADATA_FEATURE_TYPES_NOT_TARGET"),
        ],
        init=False,
        repr=False,
    )
    _fit_unused_feature_types: List[Type] = field(
        default_factory=lambda: [
            get_feature("METADATA_FEATURE_TYPES"),
            None,
            None,
            None,
        ],
        init=False,
        repr=False,
    )
    _transform_unused_feature_types: List[Type] = field(
        default_factory=lambda: [
            get_feature("METADATA_FEATURE_TYPES"),
            None,
            None,
            None,
        ],
        init=False,
        repr=False,
    )
    r_source: str = field(
        default=(Path(__file__).parent / "plot_feature_importance.R")
        .resolve()
        .as_posix(),
        init=False,
        repr=False,
    )
    main_method: str = field(default="plot_feature_importance", init=False, repr=False)

    input_columns: SelectedColumnTypes = None
    target_columns: SelectedColumnTypes = None
    sample_metadata_columns: SelectedColumnTypes = None
    sample_column: str = None
    plot_top: int = 15
    feature_meta_name: str = None
    feature_column: str = None
    cols: List[str] = None
    colHeat: List[str] = None
    dat_log: str = field(default=None, init=True, repr=False)
    show_column_names: bool = False
    scale_legend_title: str = "Value"
    column_title: str = "Samples"
    row_title: str = "Features"
    plot_title: str = "Values"

    def __post_init__(self):
        if self.dat_log == ["log2_1p", "log2"]:
            self.plot_title = "log2\n" + self.plot_title
        elif self.dat_log == ["log10_1p", "log10"]:
            self.plot_title = "log10\n" + self.plot_title


@dataclass
class FeatureImportancePlotterConfigForMetagenomics(FeatureImportancePlotterConfig):
    dat_log: str = "log2_1p"
    plot_title: str = "Abundance"
    row_title: str = "Taxa"


@dataclass
class FeatureImportancePlotterConfigForOTU(
    FeatureImportancePlotterConfigForMetagenomics
):
    row_title: str = "OTUs"


class FeatureImportancePlotter(BasePlotter):
    _config_class = FeatureImportancePlotterConfig
    config: FeatureImportancePlotterConfig

    def __init__(
        self,
        input_columns: SelectedColumnTypes = None,
        target_columns: SelectedColumnTypes = None,
        sample_metadata_columns: SelectedColumnTypes = None,
        sample_column: str = None,
        plot_top: int = Unset("15"),
        feature_meta_name: str = Unset("None"),
        feature_column: str = Unset("None"),
        cols: List[str] = Unset("None"),
        colHeat: List[str] = Unset("None"),
        dat_log: str = Unset("field(default=None, init=True, repr=False)"),
        show_column_names: bool = Unset("False"),
        scale_legend_title: str = Unset('"Value"'),
        column_title: str = Unset('"Samples"'),
        row_title: str = Unset('"Features"'),
        plot_title: str = Unset('"Values"'),
        path: Optional[str] = None,
        install_missing: bool = None,
        config: Optional[FeatureImportancePlotterConfig] = None,
    ):
        super().__init__(
            plot_top=plot_top,
            feature_meta_name=feature_meta_name,
            feature_column=feature_column,
            cols=cols,
            colHeat=colHeat,
            dat_log=dat_log,
            show_column_names=show_column_names,
            scale_legend_title=scale_legend_title,
            column_title=column_title,
            row_title=row_title,
            plot_title=plot_title,
            config=config,
            install_missing=install_missing,
        )
        self.r_caller = RCaller.from_script(self.config.r_source)
        self.r_caller.verify_r_dependencies(
            bioconductor_dependencies=["ComplexHeatmap"],
            install_missing=install_missing,
        )
        self.plotter = self.r_caller.get_method(self.config.main_method)

    def plot_dataset(
        self,
        X: "Dataset",
        feature_importances: "Dataset",
        y: "Dataset",
        sample_metadata: "Dataset" = None,
        feature_metadata: dict = None,
    ):
        from biosets import decode, get_feature_metadata, get_sample_col_name

        if feature_metadata is None:
            feature_metadata = get_feature_metadata(X)

        if self.config.sample_column is None:
            self.config.sample_column = (
                get_sample_col_name(sample_metadata)
                if sample_metadata is not None
                else get_sample_col_name(X)
            )

        self.plot_pandas(
            X=DataHandler.to_pandas(X),
            feature_importances=DataHandler.to_pandas(feature_importances),
            y=DataHandler.to_pandas(decode(y)) if y is not None else None,
            sample_metadata=DataHandler.to_pandas(sample_metadata)
            if sample_metadata is not None
            else None,
            feature_metadata=feature_metadata,
        )

    def plot_pandas(
        self,
        X: pd.DataFrame,
        feature_importances: pd.DataFrame,
        y: pd.DataFrame,
        sample_metadata: pd.DataFrame = None,
        feature_metadata: dict = None,
    ):
        if self.config.dat_log == "log2_1p":
            X = np.log2(X + 1)
        elif self.config.dat_log == "log2":
            X = np.log2(X)
        elif self.config.dat_log == "log10_1p":
            X = np.log10(X + 1)
        elif self.config.dat_log == "log10":
            X = np.log10(X)

        if feature_importances is None:
            raise ValueError("Please provide feature importances.")

        self.config.feature_column = (
            self.config.feature_column or feature_importances.columns[0]
        )
        if self.config.feature_column not in feature_importances.columns:
            raise ValueError(
                f"Feature column '{self.config.feature_column}' not found in feature "
                "importances. Please provide the column name found in both "
                "feature importances and feature metadata (if provided)."
            )

        feat_import_cols = [
            c for c in feature_importances.columns if c != self.config.feature_column
        ]

        if (
            isinstance(feature_metadata, dict)
            and feature_metadata is not None
            and len(feature_metadata) > 0
        ):
            feature_metadata = pd.DataFrame(
                list(feature_metadata.values()), index=list(feature_metadata.keys())
            )
            feature_metadata = feature_metadata.reset_index(
                names=[self.config.feature_column]
            )

        if len(feat_import_cols) > 1:
            medians = feature_importances.loc[:, feat_import_cols].median(axis=1)
            sorted_inds = np.argsort(np.abs(medians))[::-1]
            feature_importances = feature_importances.iloc[sorted_inds, :].head(
                self.config.plot_top
            )
        X = DataHandler.to_arrow(X, preserve_index=False)
        feature_importances = DataHandler.to_arrow(
            feature_importances, preserve_index=False
        )
        if y is not None:
            y = DataHandler.to_arrow(y, preserve_index=False)
        if feature_metadata is not None and len(feature_metadata) > 0:
            feature_metadata = DataHandler.to_arrow(
                feature_metadata, preserve_index=False
            )
        if sample_metadata is not None:
            sample_metadata = DataHandler.to_arrow(
                sample_metadata, preserve_index=False
            )

        self._plotter(
            X,
            y=y,
            feature_importances=feature_importances,
            sample_metadata=sample_metadata,
            feature_metadata=feature_metadata,
        )

    def _plotter(
        self,
        X,
        y,
        feature_importances,
        sample_metadata,
        feature_metadata,
    ):
        params = self.config.get_params()
        if feature_metadata is not None:
            feature_metadata_columns = DataHandler.get_column_names(feature_metadata)
            feature_meta_name = params.get("feature_meta_name")
            if feature_meta_name is not None:
                if isinstance(feature_meta_name, (str, int)):
                    feature_meta_name = [feature_meta_name]
                missing_cols = set(feature_meta_name) - set(feature_metadata_columns)
                if missing_cols:
                    raise ValueError(
                        f"Feature metadata columns {list(missing_cols)} not found in "
                        "feature metadata."
                    )

        self.plotter(
            X,
            y=y,
            feature_importances=feature_importances,
            sample_metadata=sample_metadata,
            feature_metadata=feature_metadata,
            **params,
        )

    def plot(
        self,
        X,
        feature_importances,
        y=None,
        sample_metadata=None,
        feature_metadata: dict = None,
        input_columns: SelectedColumnTypes = None,
        target_columns: SelectedColumnTypes = None,
        sample_metadata_columns: SelectedColumnTypes = None,
        sample_column: str = None,
        plot_top: int = Unset("15"),
        feature_meta_name: str = Unset("None"),
        feature_column: str = Unset("None"),
        cols: List[str] = Unset("None"),
        colHeat: List[str] = Unset("None"),
        dat_log: str = Unset("field(default=None, init=True, repr=False)"),
        show_column_names: bool = Unset("False"),
        scale_legend_title: str = Unset('"Value"'),
        column_title: str = Unset('"Samples"'),
        row_title: str = Unset('"Features"'),
        plot_title: str = Unset('"Values"'),
        path: str = None,
        device: str = "pdf",
        fingerprint: str = None,
        unused_columns: SelectedColumnTypes = None,
        raise_if_missing: bool = True,
        show=True,
    ):
        # feature_importances are not selected by columns so its None
        self.config._input_columns = self._set_input_columns_and_arity(
            input_columns, None, target_columns, sample_metadata_columns
        )
        if (
            feature_importances is not None
            and DataHandler.get_shape(feature_importances)[0] == 0
        ):
            raise ValueError("Feature importances is empty.")
        self._plot(
            X,
            feature_importances,
            y,
            sample_metadata,
            feature_metadata=feature_metadata,
            plot_top=plot_top,
            feature_meta_name=feature_meta_name,
            sample_column=sample_column,
            feature_column=feature_column,
            cols=cols,
            colHeat=colHeat,
            dat_log=dat_log,
            show_column_names=show_column_names,
            scale_legend_title=scale_legend_title,
            column_title=column_title,
            row_title=row_title,
            plot_title=plot_title,
            show=show,
            path=path,
            device=device,
            fingerprint=fingerprint,
            unused_columns=unused_columns,
            raise_if_missing=raise_if_missing,
        )
